#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import fnmatch
import importlib
import json
import os
import re
import shutil
import subprocess
import sys
import textwrap
import uuid
from enum import Enum
from pathlib import Path
from typing import Union
from urllib.parse import urlparse
import fsspec
import numpy as np
import oci
import pandas as pd
import pkg_resources
import python_jsonschema_objects as pjs
import yaml

from ads.common import logger, utils
from ads.common import auth as authutil
from ads.common import oci_client as oc
from ads.common.data import ADSData
from ads.common.model_info_extractor_factory import ModelInfoExtractorFactory
from ads.common.model_introspect import (
    TEST_STATUS,
    Introspectable,
    ModelIntrospect,
    IntrospectionNotPassed,
)
from ads.common.model_metadata import (
    METADATA_SIZE_LIMIT,
    MetadataSizeTooLarge,
    MetadataCustomCategory,
    MetadataCustomKeys,
    ModelCustomMetadata,
    ModelCustomMetadataItem,
    ModelTaxonomyMetadata,
    MetadataTaxonomyKeys,
    UseCaseType,
    InvalidObjectStoragePath,
)
from ads.config import (
    NB_SESSION_OCID,
    JOB_RUN_OCID,
    PROJECT_OCID,
    NB_SESSION_COMPARTMENT_OCID,
)
import ads.dataset.factory as factory
from ads.feature_engineering.schema import Schema, SchemaSizeTooLarge

from git import InvalidGitRepositoryError, Repo
from oci.data_science.models import ModelProvenance
from IPython.core.display import display

try:
    from yaml import CDumper as dumper
    from yaml import CLoader as loader
except:
    from yaml import Dumper as dumper
    from yaml import Loader as loader

MODEL_ARTIFACT_VERSION = "3.0"
INPUT_SCHEMA_FILE_NAME = "input_schema.json"
OUTPUT_SCHEMA_FILE_NAME = "output_schema.json"

_TRAINING_RESOURCE_OCID = JOB_RUN_OCID or NB_SESSION_OCID


class InvalidDataType(Exception):
    """Invalid Data Type."""

    pass


SAMPLE_RUNTIME_YAML = f"""
MODEL_ARTIFACT_VERSION: '{MODEL_ARTIFACT_VERSION}'
MODEL_DEPLOYMENT:
INFERENCE_CONDA_ENV:
    INFERENCE_ENV_SLUG: <slug of the conda pack>
    INFERENCE_ENV_TYPE: <data_science or published>
    INFERENCE_ENV_PATH: oci://<bucket-name>@<namespace>/<prefix>/<env>.tar.gz
    INFERENCE_PYTHON_VERSION: <python version>
"""


class ConflictStrategy(object):
    IGNORE = "IGNORE"
    UPDATE = "UPDATE"
    CREATE = "CREATE"


class PACK_TYPE(Enum):
    SERVICE_PACK = "data_science"
    USER_CUSTOM_PACK = "published"


class ChangesNotCommitted(Exception):
    def __init__(self, path):
        msg = f"""
            File(s) at {path} are either dirty or untracked.
            Please commit changes and then save the model, or set `ignore_pending_changes=True`.
        """
        super().__init__(msg)


def _oci_conda_url_breakdown(inference_conda_env):
    try:
        url_parse = urlparse(inference_conda_env)
        bucket_name = url_parse.username
        namespace = url_parse.hostname
        object_name = url_parse.path.strip("/")
        return bucket_name, namespace, object_name
    except:
        raise Exception("OCI conda url path not properly configured")


class ModelArtifact(Introspectable):
    def __init__(
        self,
        artifact_dir,
        conflict_strategy=ConflictStrategy.IGNORE,
        install_libs=False,
        reload=True,
        create=False,
        progress=None,
        model_file_name="model.onnx",
        inference_conda_env=None,
        data_science_env=False,
        ignore_deployment_error=False,
        inference_python_version=None,
    ):
        """
        A class used to construct model artifacts.

        ...

        Attributes
        ----------
        artifact_dir: str
            Path to model artifacts.
        conflict_strategy: ConflictStrategy, default: IGNORE
            Determines how to handle version conflicts between the current environment and requirements of
            model artifact.
        install_libs: bool
            Determine whether will re-install the env which the model artifact was trained in.
        reload: bool
            Determine whether will reload the Model into the env.
        create: bool
            Determine whether will generate the run time yaml.
        progress:
            Update the progress bar.
        model_file_name: str
            File to determine model serilization type.
        inference_conda_env: str
            The inferred conda env.
        data_science_env: book
            Determine whether in ODSC conda env.
        ignore_deployment_error: bool
            Determine whether to turn off logging for deployment error.
            If set to True, the prepare will ignore all the errors that may impact model deployment.

        Methods
        -------
        reload(self, model_file_name=None)
            Reloads files in model artifact directory.
        verify(self, input_data)
            Verifies a model artifact directory to be published to model catalog.
        install_requirements(self, conflict_strategy=ConflictStrategy.IGNORE)
            Installs missing libraries listed in model artifact.
        populate_metadata(self, model=None, use_case_type=None)
            Extracts and populate taxonomy metadata from given model.
        save(
            self,
            display_name: str = None,
            description: str = None,
            project_id: str = None,
            compartment_id: str = None,
            training_script_path: str = None,
            ignore_pending_changes: bool = False,
            auth: dict = None,
            training_id: str = None,
            timeout: int = None,
            ignore_introspection=False,
        )
            Saves this model artifact in model catalog.
        populate_schema(
            self,
            data_sample: ADSData = None,
            X_sample: Union[list, tuple, pd.DataFrame, pd.Series, np.ndarray] = None,
            y_sample: Union[list, tuple, pd.DataFrame, pd.Series, np.ndarray] = None,
        )
            Populates input and output schema.
        introspect(self) -> pd.DataFrame
            Runs model introspection.
        """
        self.artifact_dir = (
            artifact_dir[:-1] if artifact_dir.endswith("/") else artifact_dir
        )
        self._introspect = ModelIntrospect(self)
        self.model = None
        self.score = None
        self.inference_conda_env = inference_conda_env
        self.data_science_env = data_science_env
        self.ignore_deployment_error = ignore_deployment_error
        self.metadata_taxonomy = ModelTaxonomyMetadata()
        self.metadata_custom = ModelCustomMetadata()
        self.schema_input = Schema()
        self.schema_output = Schema()
        self._serialization_format = None
        self.inference_python_version = inference_python_version

        if create:
            self.progress = progress
            if "CONDA_PREFIX" in os.environ and "NB_SESSION_OCID" in os.environ:
                self._generate_runtime_yaml(model_file_name=model_file_name)
            else:
                self._generate_empty_runtime_yaml(model_file_name=model_file_name)
            self.version = MODEL_ARTIFACT_VERSION

        # This will re-install the env which the model artifact was trained in.
        if install_libs:
            self.install_requirements(conflict_strategy=conflict_strategy)
        # This will reload the Model into the env
        if reload:
            self.reload(model_file_name=model_file_name)

    def __repr__(self):
        res = "Artifact directory: %s\n" % self.artifact_dir
        res += "Contains: %s" % str(self._get_files())
        return res

    def __getattr__(self, item):
        return getattr(self.score, item)

    def __fetch_repo_details(self, training_code_info):
        repo = Repo(".", search_parent_directories=True)
        # get repository url
        if len(repo.remotes) > 0:
            repository_url = (
                repo.remotes.origin.url
                if repo.remotes.origin in repo.remotes
                else list(repo.remotes.values())[0].url
            )
        else:
            repository_url = "file://" + repo.working_dir  # no remote repo

        # get git branch
        git_branch = format(repo.active_branch)
        # get git commit
        try:
            git_commit = format(str(repo.head.commit.hexsha))
            training_code_info.GIT_COMMIT = git_commit
        except ValueError:
            # do not set commit if there isn't any
            pass

        training_code_info.GIT_REMOTE = repository_url
        training_code_info.GIT_BRANCH = git_branch
        training_code_info.ARTIFACT_DIRECTORY = self.artifact_dir
        return repo, training_code_info

    def __fetch_training_env_details(self, training_info):
        conda_prefix = os.environ.get("CONDA_PREFIX", None)
        pack_name = "NOT FOUND"
        try:
            manifest = fetch_manifest_from_conda_location(conda_prefix)
            manifest_type = manifest["type"]
            pack_name = manifest["pack_path"] if "pack_path" in manifest else None
            slug = manifest["slug"] if "slug" in manifest else ""

            if manifest_type == PACK_TYPE.USER_CUSTOM_PACK.value:
                if os.path.exists(
                    os.path.join(os.path.expanduser("~"), "conda", "config.yaml")
                ):
                    with open(
                        (os.path.join(os.path.expanduser("~"), "conda", "config.yaml"))
                    ) as conf:
                        user_config = yaml.load(conf, Loader=yaml.FullLoader)
                    pack_bucket = user_config["bucket_info"]["name"]
                    pack_namespace = user_config["bucket_info"]["namespace"]
                else:
                    logger.warn(
                        f"Cannot resolve the bucket name and namespace details for the conda environment {conda_prefix}. "
                        f"You can set these values while saving the model or run `odsc init -b *bucket-name* -n *namespace*` from the terminal and then rerun prepare step again."
                    )
            if not manifest_type or manifest_type.lower() not in [
                PACK_TYPE.USER_CUSTOM_PACK.value,
                PACK_TYPE.SERVICE_PACK.value,
            ]:
                if not self.ignore_deployment_error:
                    raise Exception(
                        f"Manifest type unknown. Manifest Type: {manifest_type or 'None'}"
                    )
            if not pack_name:
                if manifest_type == PACK_TYPE.USER_CUSTOM_PACK.value:
                    if self.data_science_env:
                        raise Exception(
                            f"For published environments, assign the path of the environment on \
                                        object storage to the `inference_conda_env` parameter and False to the `data_science_env` parameter."
                        )
                    error_message = f"Pack destination is not known from the manifest file in {conda_prefix}. If it was cloned from another environment, consider publishing it before preparing the model artifact. If you wish to publish the enviroment later, you may do so and provide the details while saving the model. To publish run odsc publish -s {os.path.basename(conda_prefix)} in the terminal"
                    if self.ignore_deployment_error:
                        logger.warn(error_message)
                    else:
                        if not self.inference_conda_env:
                            logger.error(error_message)
                            logger.info(
                                "You may provide oci uri to the conda environment that you wish to use witin model deployment service if you do not want to publish the current training environment."
                            )
                            raise Exception(
                                f"Could not resolve the path in the object storage for the environment: {conda_prefix}"
                            )
                else:
                    logger.warn(
                        f"Could not resolve the object storage destination of {conda_prefix}. You may provide the correct pack name and object storage details during save time."
                    )
        except Exception as e:
            raise e
        training_info.TRAINING_ENV_SLUG = slug
        if manifest_type.lower() in [
            PACK_TYPE.USER_CUSTOM_PACK.value,
            PACK_TYPE.SERVICE_PACK.value,
        ]:
            training_info.TRAINING_ENV_TYPE = manifest_type
        if pack_name:
            training_info.TRAINING_ENV_PATH = pack_name
        training_info.TRAINING_PYTHON_VERSION = sys.version.split("|")[0].strip()
        return training_info

    def __environment_details(self, model_provenance):
        model_provenance.TRAINING_REGION = os.environ.get("NB_REGION", "NOT_FOUND")
        model_provenance.TRAINING_COMPARTMENT_OCID = os.environ.get(
            "NB_SESSION_COMPARTMENT_OCID", "NOT_FOUND"
        )
        model_provenance.TRAINING_RESOURCE_OCID = os.environ.get(
            "NB_SESSION_OCID", "NOT_FOUND"
        )
        model_provenance.PROJECT_OCID = os.environ.get("PROJECT_OCID", "NOT_FOUND")
        model_provenance.TENANCY_OCID = os.environ.get("TENANCY_OCID", "NOT_FOUND")
        model_provenance.USER_OCID = os.environ.get("USER_OCID", "NOT_FOUND")
        model_provenance.VM_IMAGE_INTERNAL_ID = os.environ.get("VM_ID", "VMIDNOTSET")
        return model_provenance

    def __fetch_runtime_schema__(self):
        schema = None
        with open(
            os.path.join(
                os.path.dirname(os.path.abspath(__file__)), "model_artifact_schema.json"
            )
        ) as schema_file:
            schema = json.load(schema_file)

        if not schema:
            raise Exception("Cannot load schema file for generating runtime yaml")

        builder = pjs.ObjectBuilder(schema)
        ns = builder.build_classes()
        return ns

    def _generate_empty_runtime_yaml(self, model_file_name="model.onnx"):
        logger.warning(
            "Generating runtime.yaml template. This file needs to be updated with valid values before saving to model catalog for a successful deployment"
        )
        content = SAMPLE_RUNTIME_YAML
        with open(os.path.join(self.artifact_dir, "runtime.yaml"), "w") as outfile:
            outfile.write(content)

    def _generate_runtime_yaml(self, model_file_name="model.onnx"):
        if self.progress:
            self.progress.update("Creating runtime.yaml configuration")

        ns = self.__fetch_runtime_schema__()
        training_env_info = self.__fetch_training_env_details(ns.TrainingCondaEnv())
        model_provenance = self.__environment_details(ns.ModelProvenance())
        model_provenance.TRAINING_CONDA_ENV = training_env_info
        try:
            _, training_code_info = self.__fetch_repo_details(ns.TrainingCodeInfo())
            model_provenance.TRAINING_CODE = training_code_info
        except InvalidGitRepositoryError:
            pass

        if not training_env_info.TRAINING_ENV_PATH:
            logger.warning(
                "We noticed that you did not publish the conda environment in which you trained the model. Publishing the conda environment you used to train the model is optional when you save the model to the catalog. However, publishing the environment ensures that the exact training environment can be re-used later"
            )
        inference_info = ns.InferenceCondaEnv()
        if not self.inference_conda_env:
            message = f"We give you the option to specify a different inference conda environment for model deployment purposes. By default it is assumed to be the same as the conda environment used to train the model. If you wish to specify a different environment for inference purposes, please assign the path of a published or data science conda environment to the optional parameter `inference_conda_env`. "
            if (
                training_env_info.TRAINING_ENV_TYPE
                and training_env_info.TRAINING_ENV_PATH
            ):
                logger.info(message)
                inference_info.INFERENCE_ENV_SLUG = training_env_info.TRAINING_ENV_SLUG
                inference_info.INFERENCE_ENV_TYPE = training_env_info.TRAINING_ENV_TYPE
                inference_info.INFERENCE_ENV_PATH = training_env_info.TRAINING_ENV_PATH
                inference_info.INFERENCE_PYTHON_VERSION = (
                    training_env_info.TRAINING_PYTHON_VERSION
                )
                self.conda_env = str(training_env_info.TRAINING_ENV_SLUG)
        else:
            self.conda_env = os.path.basename(str(self.inference_conda_env))
            if self.inference_conda_env.startswith("oci://"):
                inference_info.INFERENCE_ENV_PATH = self.inference_conda_env
                try:
                    bucket_name, namespace, object_name = _oci_conda_url_breakdown(
                        self.inference_conda_env
                    )
                    auth = authutil.default_signer()
                    oci_client = oc.OCIClientFactory(**auth).object_storage
                    res = oci_client.get_object(namespace, bucket_name, object_name)
                    metadata = res.data.headers["opc-meta-manifest"]
                    metadata_json = json.loads(metadata)
                    inference_info.INFERENCE_PYTHON_VERSION = metadata_json["python"]
                except:
                    if not self.inference_python_version:
                        if not training_env_info.TRAINING_PYTHON_VERSION:
                            raise Exception(
                                "Python version is not specified, please pass in python version when preparing model."
                            )
                        else:
                            logger.warning(
                                "We could not infer python version from conda pack so we are defaulting inference "
                                "python version to the training python version (which comes from sys.version)."
                            )
                            inference_info.INFERENCE_PYTHON_VERSION = (
                                training_env_info.TRAINING_PYTHON_VERSION
                            )
                    else:
                        inference_info.INFERENCE_PYTHON_VERSION = (
                            self.inference_python_version
                        )
            else:
                pass
        model_deployment_info = None
        if inference_info.INFERENCE_ENV_PATH:
            model_deployment_info = ns.ModelDeployment()
            model_deployment_info.INFERENCE_CONDA_ENV = inference_info

        if (
            not self.inference_conda_env
            and not self.data_science_env
            and inference_info.INFERENCE_ENV_TYPE == PACK_TYPE.SERVICE_PACK.value
            and training_env_info.TRAINING_ENV_PATH == inference_info.INFERENCE_ENV_PATH
        ):
            error_message = f"The inference environment {training_env_info.TRAINING_ENV_SLUG} may have undergone changes over the course of development. You can choose to publish the current environment or set data_science_env to True in the prepare api"
            if not self.ignore_deployment_error:
                raise Exception(error_message)
            else:
                logger.warning(error_message)

        if not inference_info.INFERENCE_ENV_PATH and not self.inference_conda_env:
            error_message = f"Missing information about conda environment to use during inference in model deployment service. You can provide the oci uri of the pack that you wish to use as inference in the paramenter inference_conda_env or publish your environment by following the instruction on Environment Explorer and then run the prepare method again "
            if not self.ignore_deployment_error:
                raise Exception(error_message)
            else:
                logger.warn(error_message)

        if model_deployment_info:
            self._runtime_info = ns.ModelArtifactSchema(
                MODEL_ARTIFACT_VERSION=MODEL_ARTIFACT_VERSION,
                MODEL_PROVENANCE=model_provenance,
                MODEL_DEPLOYMENT=model_deployment_info,
            )
        else:
            self._runtime_info = ns.ModelArtifactSchema(
                MODEL_ARTIFACT_VERSION=MODEL_ARTIFACT_VERSION,
                MODEL_PROVENANCE=model_provenance,
            )

        with open(os.path.join(self.artifact_dir, "runtime.yaml"), "w") as outfile:
            outfile.write("# Model runtime environment\n")
            yaml.dump(self._runtime_info.as_dict(), outfile, default_flow_style=False)

    def reload(self, model_file_name: str = None):
        """
        Reloads files in model artifact directory.

        Parameters
        ----------
        model_file_name: str
            The model file name.
        """
        spec = importlib.util.spec_from_file_location(
            "score%s" % uuid.uuid4(), os.path.join(self.artifact_dir, "score.py")
        )
        score = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(score)
        self.score = score
        if os.path.exists(os.path.join(self.artifact_dir, "runtime.yaml")):
            if model_file_name:
                self.model = self.score.load_model(model_file_name)
            else:
                self.model = self.score.load_model()
            with open(os.path.join(self.artifact_dir, "runtime.yaml")) as runtime_file:
                runtime = yaml.load(runtime_file, Loader=yaml.FullLoader)
            self.version = runtime["MODEL_ARTIFACT_VERSION"]
            try:
                self.VM_ID = runtime["MODEL_PROVENANCE"]["VM_IMAGE_INTERNAL_ID"]
            except KeyError:
                self.VM_ID = None
            try:
                self.conda_env = runtime["MODEL_PROVENANCE"]["TRAINING_CONDA_ENV"][
                    "TRAINING_ENV_SLUG"
                ]
            except KeyError:
                self.conda_env = None
        elif os.path.exists(os.path.join(self.artifact_dir, "ds-runtime.yaml")):
            self.model = self.score.load_model()
            with open(
                os.path.join(self.artifact_dir, "ds-runtime.yaml")
            ) as runtime_file:
                runtime = yaml.load(runtime_file, Loader=yaml.FullLoader)
            self.version = "1.0"
            self.VM_ID = None  # get ads/mlx version?
            self.conda_env = runtime["conda-env"]
        else:
            self.model = self.score.load_model()
            self.version = "0.0"
            self.VM_ID = "UNKNOWN"
            self.conda_env = "base"
            # raise FileNotFoundError(os.path.join(self.artifact_dir, 'runtime.yaml'))
        # __pycache__ was created during model_artifact.reload() above
        if os.path.exists(os.path.join(self.artifact_dir, "__pycache__")):
            shutil.rmtree(
                os.path.join(self.artifact_dir, "__pycache__"), ignore_errors=True
            )
        # extract model serialization format as part of custom metadata
        if model_file_name:
            self._serialization_format = self._extract_model_serialization_format(
                model_file_name
            )
            if (
                MetadataCustomKeys.MODEL_SERIALIZATION_FORMAT
                in self.metadata_custom.keys
            ):
                self.metadata_custom[
                    MetadataCustomKeys.MODEL_SERIALIZATION_FORMAT
                ].value = self._serialization_format
            else:
                self.metadata_custom.add(
                    key=MetadataCustomKeys.MODEL_SERIALIZATION_FORMAT,
                    value=self._serialization_format,
                    description="The model serialization format",
                    category=MetadataCustomCategory.TRAINING_PROFILE,
                )

    @staticmethod
    def _extract_model_serialization_format(model_file_name):
        return os.path.splitext(model_file_name)[1][1:]

    def verify(self, input_data):
        """
        Verifies a model artifact directory to be published to model catalog.

        Parameters
        ----------
        input_data : str, dict, BytesIO stream
            Data to be passed into the deployed model. It can be of type json (str), a dict object, or a BytesIO stream.
            All types get converted into a UTF-8 encoded BytesIO stream and is then sent to the handler.
            Any data handling past there is done in func.py (which the user can edit). By default it looks for data
            under the keyword "input", and returns data under teh keyword "prediction"

        Returns
        -------
        output_data : the resulting prediction, formatted in the same way as the input

        Example
         --------
         input_dict = {"input": train.X[:3].to_dict()}
         model_artifact.verify(input_dict)

         * returns {"prediction": [30/4, 24.8, 30.7]} *
        """

        # Fake Context obj created for Fn Handler
        class FakeCtx:
            def SetResponseHeaders(self, headers, status_code):
                return

        ctx = FakeCtx()
        from io import BytesIO

        if type(input_data) == str:
            data = BytesIO(input_data.encode("UTF-8"))
            data_type = "json"
        elif type(input_data) == dict:
            from json import dumps

            data = BytesIO(dumps(input_data).encode("UTF-8"))
            data_type = "dict"
        elif isinstance(type(input_data), type(BytesIO)):
            data = input_data
            data_type = "BytesIO"
        else:
            raise TypeError

        sys_path = sys.path.copy()
        try:
            if self.version.split(".")[0] not in ["0", "1"]:
                sys.path.insert(0, self.artifact_dir)
            else:
                sys.path.insert(0, os.path.join(self.artifact_dir, "fn-model"))
            import func

            resp = func.handler(ctx, data)
            output_json = resp.body()
        finally:
            # Reset in case func.py messes with it
            sys.path = sys_path

        if data_type == "json":
            return output_json
        output_bstream = BytesIO(resp.body().encode("UTF-8"))
        if data_type == "BytesIO":
            return output_bstream
        else:
            from json import load

            return load(output_bstream)

    def save(
        self,
        display_name: str = None,
        description: str = None,
        project_id: str = None,
        compartment_id: str = None,
        training_script_path: str = None,
        ignore_pending_changes: bool = False,
        auth: dict = None,
        training_id: str = None,
        timeout: int = None,
        ignore_introspection=True,
    ):
        """
        Saves this model artifact in model catalog.

        Parameters
        ----------
        display_name : str, optional
            Model display name
        description : str, optional
            Description for the model
        project_id : str, optional
            OCID of models' project
            If None, the default project ID `config.PROJECT_OCID` would be used
        compartment_id : str, optional
            OCID of model's compartment
            If None, the default compartment ID `config.NB_SESSION_COMPARTMENT_OCID` would be used
        training_script_path : str, optional
            The training script path is either relative to the working directory,
            or an absolute path.
        ignore_pending_changes : bool, default: False
            If True, ignore uncommitted changes and use the current head commit for provenance metadata
            This argument is used only when the function is called from a script in git managed directory
        auth: dict
            Default is None. The default authetication is set using `ads.set_auth` API. If you need to override the
            default, use the `ads.common.auth.api_keys` or `ads.common.auth.resource_principal` to create appropriate
            authentication signer and kwargs required to instantiate DataScienceClient object
        training_id: str, optional
            The training OCID for model.
        timeout: int, optional
            The connection timeout in seconds for the client.
            The default value for connection timeout is 10 seconds.
        ignore_introspection: bool, optional
            Determine whether to ignore the result of model introspection or not.
            If set to True, the save will ignore all model introspection errors.

        Examples
        ________
        >>> from ads.common.model_artifact import ModelArtifact
        >>> from ads.config import NB_SESSION_OCID

        >>> # Getting auth details.
        >>> # If you are using API keys
        >>> auth=ads.common.auth.api_keys()

        >>> # If you are using resource principal
        >>> auth=ads.common.auth.resource_principal()

        >>> # If you have set the auth type using ads.set_auth()
        >>> auth=ads.common.auth.default_signer()

        >>> # Preparing model artifacts
        >>> model_artifact = prepare_generic_model(
        ...     "path_to_model_artifacts",
        ...     force_overwrite=True,
        ...     data_science_env=True,
        ...     model=gamma_reg_model,
        ... )

        >>> # Saving model to the model catalog
        >>> model_artifact.save(
        ...     project_id=PROJECT_ID,
        ...     compartment_id=COMPARTMENT,
        ...     display_name="RF Classifier 2",
        ...     description="A sample Random Forest classifier",
        ...     ignore_pending_changes=True,
        ...     auth=auth,
        ...     training_id=NB_SESSION_OCID,
        ...     timeout=6000,
        ...     ignore_introspection = True
        ... )
        """
        if timeout and not isinstance(timeout, int):
            raise TypeError("Timeout must be an integer.")

        runtime_yaml_file = os.path.join(self.artifact_dir, "runtime.yaml")
        if os.path.exists(runtime_yaml_file):
            with open(runtime_yaml_file, "r") as mfile:
                runtime_prep_info = yaml.load(mfile, Loader=yaml.FullLoader)
                # runtime_info['pack-info'] = deployment_pack_info
        else:
            runtime_prep_info = "{}"
        ns = self.__fetch_runtime_schema__()
        runtime_info = ns.ModelArtifactSchema().from_json(json.dumps(runtime_prep_info))

        training_code_info = self._training_code_info(
            ns, training_script_path, ignore_pending_changes
        )
        if not training_id:
            training_id = _TRAINING_RESOURCE_OCID
        model_provenance_metadata = ModelProvenance(
            repository_url=str(training_code_info.GIT_REMOTE),
            git_branch=str(training_code_info.GIT_BRANCH),
            git_commit=str(training_code_info.GIT_COMMIT),
            script_dir=str(training_code_info.ARTIFACT_DIRECTORY),
            training_script=str(training_code_info.TRAINING_SCRIPT),
            training_id=training_id,
        )
        if getattr(runtime_info, "MODEL_PROVENANCE", None):
            runtime_info.MODEL_PROVENANCE.TRAINING_CODE = training_code_info

        logger.info(model_provenance_metadata)

        # handle the case where project_id and/or compartment_id is not specified by the user
        if not project_id and not PROJECT_OCID:
            raise ValueError("The `project_id` must be provided.")

        if not compartment_id and not NB_SESSION_COMPARTMENT_OCID:
            raise ValueError("The `compartment_id` must be provided.")

        if os.path.exists(os.path.join(self.artifact_dir, "__pycache__")):
            shutil.rmtree(
                os.path.join(self.artifact_dir, "__pycache__"), ignore_errors=True
            )
        self.metadata_custom._add(
            ModelCustomMetadataItem(
                key=MetadataCustomKeys.MODEL_ARTIFACTS,
                value=textwrap.shorten(
                    ", ".join(self._get_files()), 255, placeholder="..."
                ),
                description="The list of files located in artifacts folder",
                category=MetadataCustomCategory.TRAINING_ENV,
            ),
            replace=True,
        )

        client_auth = auth if auth else authutil.default_signer()

        if timeout:
            if not client_auth.get("client_kwargs"):
                client_auth["client_kwargs"] = {}
            client_auth["client_kwargs"]["timeout"] = timeout

        self._validate_metadata()
        self._validate_schema()

        with open(runtime_yaml_file, "w") as mfile:
            yaml.dump(runtime_info.as_dict(), mfile, Dumper=dumper)

        if not ignore_introspection:
            self._introspect()
            if self._introspect.status == TEST_STATUS.NOT_PASSED:
                msg = (
                    "Model introspection not passed. "
                    "See the table below for "
                    "more detailed information and follow the messages to fix it, "
                    "otherwise you might come across errors during model deployment. "
                    "To save model artifacts ignoring "
                    "introspection use .save(ignore_introspection=True...)."
                )
                logger.error(msg)
                display(self._introspect.to_dataframe())
                raise (IntrospectionNotPassed())

        try:
            from ads.catalog.model import ModelCatalog

            return ModelCatalog(
                compartment_id=compartment_id,
                ds_client_auth=client_auth,
                identity_client_auth=client_auth,
            ).upload_model(
                self,
                provenance_metadata=model_provenance_metadata,
                display_name=display_name,
                description=description,
                project_id=project_id,
            )
        except oci.exceptions.RequestException as e:
            if "The write operation timed out" in str(e):
                logger.error(
                    "The save operation timed out. Try to set a longer timeout(e.g. save(timeout=600, ...))."
                )

    def _validate_schema(self):
        if not self._validate_schema_size(self.schema_input, INPUT_SCHEMA_FILE_NAME):
            self.schema_input.to_json_file(
                os.path.join(self.artifact_dir, INPUT_SCHEMA_FILE_NAME)
            )
        if not self._validate_schema_size(self.schema_output, OUTPUT_SCHEMA_FILE_NAME):
            self.schema_output.to_json_file(
                os.path.join(self.artifact_dir, OUTPUT_SCHEMA_FILE_NAME)
            )
        self.schema_input.validate_schema()
        self.schema_output.validate_schema()

    def _validate_metadata(self):
        self.metadata_custom.validate()
        self.metadata_taxonomy.validate()
        total_size = self.metadata_custom.size() + self.metadata_taxonomy.size()
        if total_size > METADATA_SIZE_LIMIT:
            raise MetadataSizeTooLarge(total_size)
        return True

    def _training_code_info(
        self, ns, training_script_path=None, ignore_pending_changes=False
    ):
        try:
            repo, training_code_info = self.__fetch_repo_details(ns.TrainingCodeInfo())
        except InvalidGitRepositoryError:
            repo = None
            training_code_info = ns.TrainingCodeInfo()
        if training_script_path is not None:
            if not os.path.exists(training_script_path):
                logger.warning(
                    f"Training script {os.path.abspath(training_script_path)} does not exists."
                )
            else:
                training_script = os.path.abspath(training_script_path)
                self._assert_path_not_dirty(
                    training_script_path, repo, ignore_pending_changes
                )
                training_code_info.TRAINING_SCRIPT = training_script

        self._assert_path_not_dirty(self.artifact_dir, repo, ignore_pending_changes)
        training_code_info.ARTIFACT_DIRECTORY = os.path.abspath(self.artifact_dir)

        return training_code_info

    def _assert_path_not_dirty(self, path, repo, ignore):
        if repo is not None and not ignore:
            path_abs = os.path.abspath(path)
            if os.path.commonpath([path_abs, repo.working_dir]) == repo.working_dir:
                path_relpath = os.path.relpath(path_abs, repo.working_dir)
                if repo.is_dirty(path=path_relpath) or any(
                    [
                        os.path.commonpath([path_relpath, untracked]) == path_relpath
                        for untracked in repo.untracked_files
                    ]
                ):
                    raise ChangesNotCommitted(path_abs)

    def install_requirements(self, conflict_strategy=ConflictStrategy.IGNORE):
        """
        Installs missing libraries listed in model artifact.

        Parameters
        ----------
        conflict_strategy : ConflictStrategy, default: IGNORE
            Update the conflicting dependency to the version required by the model artifact.
            Valid values: "IGNORE" or ConflictStrategy.IGNORE, "UPDATE" or ConflictStrategy.UPDATE
            IGNORE: Use the installed version in  case of a conflict
            UPDATE: Force update dependency to the version required by model artifact in case of conflict
        """
        importlib.reload(pkg_resources)
        from pkg_resources import DistributionNotFound, VersionConflict

        if self.version.split(".")[0] not in ["0", "1"] and os.path.exists(
            Path(os.path.join(self.artifact_dir), "requirements.txt")
        ):
            requirements = (
                Path(os.path.join(self.artifact_dir), "requirements.txt")
                .read_text()
                .strip()
                .split("\n")
            )
        elif self.version.split(".")[0] in ["0", "1"] and Path(
            os.path.join(self.artifact_dir), "ds-requirements.txt"
        ):
            requirements = (
                Path(os.path.join(self.artifact_dir), "ds-requirements.txt")
                .read_text()
                .strip()
                .split("\n")
            )
        else:
            raise FileNotFoundError(
                "Could not find requirements file. Please install necessary libraries and "
                "re-construct the model artifact with install_libs=False"
            )

        version_conflicts = {}
        for requirement in requirements:
            try:
                pkg_resources.require(requirement)
            except VersionConflict as vc:
                if conflict_strategy == ConflictStrategy.UPDATE:
                    pip_install("%s%s" % (vc.req.name, vc.req.specifier), "-U")
                elif conflict_strategy == ConflictStrategy.IGNORE:
                    version_conflicts[
                        "%s==%s" % (vc.dist.key, vc.dist.parsed_version)
                    ] = "%s%s" % (vc.req.name, vc.req.specifier)
            except DistributionNotFound as dnf:
                pip_install(requirement)
                # distributions_not_found.add('%s%s' % (dnf.req.name, dnf.req.specifier))
        if len(version_conflicts) > 0:
            print(
                "\033[93m"
                + str(VersionConflictWarning(version_conflicts=version_conflicts))
                + "\033[0m"
            )

    def _get_files(self):
        if os.path.exists(os.path.join(self.artifact_dir, ".model-ignore")):
            ignore_patterns = (
                Path(os.path.join(self.artifact_dir), ".model-ignore")
                .read_text()
                .strip()
                .split("\n")
            )
        else:
            ignore_patterns = []
        file_names = []
        for root, dirs, files in os.walk(self.artifact_dir):
            for name in files:
                file_names.append(os.path.join(root, name))
            for name in dirs:
                file_names.append(os.path.join(root, name))

        for ignore in ignore_patterns:
            if not ignore.startswith("#") and ignore.strip() != "":
                matches = []
                for file_name in file_names:
                    if ignore.endswith("/"):
                        ignore = ignore[:-1] + "*"
                    if not re.search(
                        fnmatch.translate("/%s" % ignore.strip()), file_name
                    ):
                        matches.append(file_name)
                file_names = matches
        return [
            matched_file[len(self.artifact_dir) + 1 :] for matched_file in file_names
        ]

    def _save_data_from_memory(
        self,
        prefix: str,
        train_data: Union[pd.DataFrame, list, np.ndarray],
        validation_data: Union[pd.DataFrame, list, np.ndarray] = None,
        train_data_name: str = "train.csv",
        validation_data_name: str = "validation.csv",
        storage_options: dict = None,
        **kwargs,
    ):
        """
        Save the data to object storage.
        return [
            matched_file[len(self.artifact_dir) + 1 :] for matched_file in file_names
        ]

        Parameters
        ----------
        prefix: str
            The object storage prefix to store the data.
            e.g. oci://bucket_name@namespace/folder_name
        train_data: Union[pd.DataFrame, list, np.ndarray].
            The training data to be stored.
        validation_data: Union[pd.DataFrame, list, np.ndarray]. Default None
            The validation data to be stored.
        train_data_name: str. Default 'train.csv'.
            File name for train data to be stored on the cloud. train_data is
            stored at prefix/train_data_name.
        validation_data_name: str. Default 'train.csv'.
            File name for validation data to be stored on the cloud.
            validation_data is stored at prefix/validation_data_name.
        storage_options: dict. Default None
            Parameters passed on to the backend filesystem class.
            Defaults to `storage_options` set using `DatasetFactory.set_default_storage()`.

        Returns
        -------
        None
            Nothing.
        Examples
        ________
        >>> from ads.common.model_artifact import ModelArtifact
        >>> import ocifs
        >>> import oci
        >>> storage_options = {"config": oci.config.from_file(os.path.join("~/.oci", "config"))}
        >>> storage_options
        {'log_requests': False,
            'additional_user_agent': '',
            'pass_phrase': None,
            'user': 'ocid5.user.oc1..aaaaaaaab3geixlk***********************',
            'fingerprint': '05:15:2b:b1:46:8a:32:ec:e2:69:5b:32:01:**:**:**)',
            'tenancy': 'ocid5.tenancy.oc1..aaaaaaaag*************************',
            'region': 'us-ashburn-1',
            'key_file': '/home/datascience/.oci/oci_api_key.pem'}
        >>> path_to_generic_model_artifact = tempfile.mkdtemp()
        >>> df = pd.DataFrame([[1, 2], [2, 3], [3, 4], [4, 3]])
        >>> generic_model_artifact = prepare_generic_model(path_to_generic_model_artifact,
        ...                                   force_overwrite=True, data_science_env=True,
        ...                                   ignore_deployment_error=True)
        >>> generic_model_artifact._save_data_from_memory(prefix = 'oci://bucket_name@namespace/folder_name',
        ... train_data=df, storage_options=storage_options)
        """
        if not re.match(r"oci://*@*", prefix):
            raise InvalidObjectStoragePath(
                "`prefix` is not valid. It must have the following pattern 'oci://bucket_name@namespace/key'."
            )
        if not storage_options:
            storage_options = factory.default_storage_options
        if not storage_options:
            storage_options = {"config": {}}

        self._save_from_memory(
            train_data, prefix, train_data_name, storage_options, "training", **kwargs
        )

        if validation_data is not None:
            self._save_from_memory(
                validation_data,
                prefix,
                validation_data_name,
                storage_options,
                "validation",
                **kwargs,
            )

    def _save_data_from_file(
        self,
        prefix: str,
        train_data_path: str = None,
        validation_data_path: str = None,
        storage_options: dict = None,
        **kwargs,
    ):
        """
        Save the data to object storage.

        Parameters
        ----------
        prefix: str
            The object storage prefix to store the data. When `train_data_path` or
            `validation_data_path` are provided, they are stored under this prefix
            with their original file names. If the data are already stored on the object
            storage, you can provide the path of the If none of the local data path is provided,
            no data is `prefix` is saved in the custom metadata.
        train_data_path: str. Default None.
            Local path for training data.
        validation_data_path: str. Default None.
            Local path for validation data.
        storage_options: dict. Default None
            Parameters passed on to the backend filesystem class.
            Defaults to `storage_options` set using `DatasetFactory.set_default_storage()`.
        Keyword Arguments
        _________________
        data_type:
            Can be either `training` or `validation`. Only used when
            your data are already stored on the cloud and you only want to
            record the info of the data path under `metadata_custom`.
            Pass the prefix of your data and `data_type` to indicate whether
            this data is of `training` type or `validation` type.
        Note
        ____
        In the case when your data are already stored on the cloud and you only want to
        record the info of the data path under `metadata_custom`, you can pass the
        path where your data are stored on the cloud to prefix and it records this information
        to `metadata_custom`. `storage_options` is needed in this case.

        Returns
        -------
        None
            Nothing.

        Examples
        ________
        >>> from ads.common.model_artifact import ModelArtifact
        >>> import ocifs
        >>> import oci
        >>> storage_options = {"config": oci.config.from_file(os.path.join("~/.oci", "config"))}
        >>> storage_options
        {'log_requests': False,
            'additional_user_agent': '',
            'pass_phrase': None,
            'user': 'ocid5.user.oc1..aaaaaaaab3geixlk***********************',
            'fingerprint': '05:15:2b:b1:46:8a:32:ec:e2:69:5b:32:01:**:**:**)',
            'tenancy': 'ocid5.tenancy.oc1..aaaaaaaag*************************',
            'region': 'us-ashburn-1',
            'key_file': '/home/datascience/.oci/oci_api_key.pem'}
        >>> path_to_generic_model_artifact = tempfile.mkdtemp()
        >>> generic_model_artifact = prepare_generic_model(path_to_generic_model_artifact,
        ...                                   force_overwrite=True, data_science_env=True,
        ...                                   ignore_deployment_error=True)
        >>> generic_model_artifact._save_data_from_file(oci_storage_path = 'oci://bucket_name@namespace/folder_name',
        ... train_data_path = '~/orcl_attrition*.csv', storage_options=storage_options)
        """
        if not re.match(r"oci://*@*", prefix):
            raise InvalidObjectStoragePath(
                "`prefix` is not valid. It must have the following pattern 'oci://bucket_name@namespace/key'."
            )
        if not storage_options:
            storage_options = factory.default_storage_options
        if not storage_options:
            storage_options = {"config": {}}

        if train_data_path is not None:
            assert isinstance(train_data_path, str), "A path to the data is required."
            self._save_from_local_file(
                prefix=prefix,
                file_path=train_data_path,
                storage_options=storage_options,
                data_type="training",
            )
        if validation_data_path is not None:
            assert isinstance(
                validation_data_path, str
            ), "A path to the data is required."
            self._save_from_local_file(
                prefix=prefix,
                file_path=validation_data_path,
                storage_options=storage_options,
                data_type="validation",
            )
        if train_data_path is None and validation_data_path is None:
            data_type = kwargs.get("data_type", "training")
            if data_type not in ("training", "validation"):
                InvalidDataType(
                    "`data_type` is not supported. Choose 'training' or 'validation'."
                )
            self._save_data_path(prefix, data_type=data_type)

    def _populate_metadata_taxonomy(self, model=None, use_case_type=None):
        """Extracts and populate taxonomy metadata from the given model.

        Parameters
        ----------
        model: [sklearn, xgboost, lightgbm, automl, keras]
            The model object.
        use_case_type: str
            The use case type of the model
        Returns
        -------
        None
            Nothing.

        Raises
        ------
        ValueError: When model not provided.
        """
        if use_case_type and use_case_type not in UseCaseType:
            raise ValueError(
                f"Invalid value of `UseCaseType`. Choose from {UseCaseType.values()}."
            )

        self.metadata_taxonomy[MetadataTaxonomyKeys.USE_CASE_TYPE].value = use_case_type
        if model is not None:
            map = ModelInfoExtractorFactory.extract_info(model)
            if map is not None:
                self.metadata_taxonomy._populate_from_map(map)
            if (
                self.metadata_taxonomy[MetadataTaxonomyKeys.HYPERPARAMETERS].size()
                > METADATA_SIZE_LIMIT
            ):
                logger.warn(
                    f"The model hyperparameters are larger than `{METADATA_SIZE_LIMIT}` "
                    "bytes and cannot be stored as model catalog metadata. It will be saved to "
                    f"{self.artifact_dir}/hyperparameters.json and removed from the metadata."
                )

                self.metadata_taxonomy[
                    MetadataTaxonomyKeys.HYPERPARAMETERS
                ].to_json_file(self.artifact_dir)
                self.metadata_taxonomy[MetadataTaxonomyKeys.HYPERPARAMETERS].update(
                    value=None
                )

    def _populate_metadata_custom(self):
        """Extracts custom metadata info from model artifact.

        Parameters
        ----------
        None
            Nothing

        Returns
        -------
        None
            Nothing
        """
        model_metadata_items = []

        model_metadata_items.append(
            ModelCustomMetadataItem(
                key=MetadataCustomKeys.CONDA_ENVIRONMENT,
                value=self.conda_env if hasattr(self, "conda_env") else None,
                description="The conda env where model was trained",
                category=MetadataCustomCategory.TRAINING_ENV,
            )
        )
        try:
            env_type = (
                self._runtime_info.MODEL_DEPLOYMENT.INFERENCE_CONDA_ENV.INFERENCE_ENV_TYPE._value
            )
        except:
            env_type = None
        try:
            slug_name = (
                self._runtime_info.MODEL_DEPLOYMENT.INFERENCE_CONDA_ENV.INFERENCE_ENV_SLUG._value
            )
        except:
            slug_name = None
        try:
            env_path = (
                self._runtime_info.MODEL_DEPLOYMENT.INFERENCE_CONDA_ENV.INFERENCE_ENV_PATH._value
            )
        except:
            env_path = None

        model_metadata_items.append(
            ModelCustomMetadataItem(
                key=MetadataCustomKeys.ENVIRONMENT_TYPE,
                value=env_type,
                description="The env type, could be published conda or datascience conda",
                category=MetadataCustomCategory.TRAINING_ENV,
            )
        )
        model_metadata_items.append(
            ModelCustomMetadataItem(
                key=MetadataCustomKeys.SLUG_NAME,
                value=slug_name,
                description="The slug name of the conda env where model was trained",
                category=MetadataCustomCategory.TRAINING_ENV,
            )
        )
        model_metadata_items.append(
            ModelCustomMetadataItem(
                key=MetadataCustomKeys.CONDA_ENVIRONMENT_PATH,
                value=env_path,
                description="The oci path of the conda env where model was trained",
                category=MetadataCustomCategory.TRAINING_ENV,
            )
        )
        model_metadata_items.append(
            ModelCustomMetadataItem(
                key=MetadataCustomKeys.MODEL_ARTIFACTS,
                value=textwrap.shorten(
                    ", ".join(self._get_files()), 255, placeholder="..."
                ),
                description="The list of files located in artifacts folder",
                category=MetadataCustomCategory.TRAINING_ENV,
            )
        )
        model_metadata_items.append(
            ModelCustomMetadataItem(
                key=MetadataCustomKeys.MODEL_SERIALIZATION_FORMAT,
                value=self._serialization_format,
                description="The model serialization format",
                category=MetadataCustomCategory.TRAINING_PROFILE,
            )
        )
        model_metadata_items.append(
            ModelCustomMetadataItem(
                key=MetadataCustomKeys.CLIENT_LIBRARY,
                value="ADS",
                description="",
                category=MetadataCustomCategory.OTHER,
            )
        )
        self.metadata_custom._add_many(model_metadata_items, replace=True)

    def populate_metadata(self, model=None, use_case_type=None):
        """Extracts and populate taxonomy metadata from given model.

        Parameters
        ----------
        model: [sklearn, xgboost, lightgbm, automl, keras]
            The model object

        use_case_type:
            The use case type of the model

        Returns
        -------
        None
            Nothing.
        """
        if model is None and self.metadata_taxonomy["Algorithm"].value is None:
            logger.info(
                "To auto-extract taxonomy metadata the model must be provided. Supported models: automl, keras, lightgbm, pytorch, sklearn, tensorflow, and xgboost."
            )
        if use_case_type is None:
            use_case_type = self.metadata_taxonomy[
                MetadataTaxonomyKeys.USE_CASE_TYPE
            ].value
        self._populate_metadata_taxonomy(model, use_case_type)
        self._populate_metadata_custom()

    def _save_from_memory(
        self,
        data,
        prefix,
        data_file_name,
        storage_options,
        data_type="training",
        **kwargs,
    ):
        """
        Save the data to object storage.
        """
        oci_storage_path = os.path.join(prefix, data_file_name)
        if isinstance(data, np.ndarray) or isinstance(data, list):
            data = pd.DataFrame(data)
            data.to_csv(oci_storage_path, storage_options=storage_options, **kwargs)
        elif isinstance(data, pd.Series) or isinstance(data, pd.DataFrame):
            data.to_csv(oci_storage_path, storage_options=storage_options, **kwargs)
        elif isinstance(data, ADSData):
            data = pd.concat([data.X, data.y], axis=1)
            data.to_csv(oci_storage_path, storage_options=storage_options, **kwargs)
        else:
            raise NotImplementedError(
                f"`{type(data)}` is not supported. Use a pandas DataFrame."
            )

        self._save_data_path(oci_storage_path, data_type)
        self._save_data_shape(data, data_type)

    def _save_from_local_file(
        self, prefix, file_path, storage_options, data_type="training"
    ):
        """Save local file to object storage."""
        file_path = os.path.expanduser(file_path)
        import glob

        if len(glob.glob(file_path)) == 0:
            raise FileExistsError(f"No files were found in `{file_path}`.")
        oci_storage_paths = []
        with fsspec.open_files(file_path, mode="r") as fhs:
            for fh in fhs:
                oci_storage_path = os.path.join(prefix, os.path.basename(fh.name))
                with fsspec.open(
                    oci_storage_path,
                    mode="w",
                    **(storage_options),
                ) as f:
                    f.write(fh.read())
                oci_storage_paths.append(oci_storage_path)
                self._save_file_size(
                    os.path.join(os.path.dirname(file_path), os.path.basename(fh.name)),
                    data_type,
                )
        self._save_data_path(",  ".join(oci_storage_paths), data_type)

    def _save_data_path(self, oci_storage_path, data_type):

        key = (
            MetadataCustomKeys.TRAINING_DATASET
            if data_type == "training"
            else MetadataCustomKeys.VALIDATION_DATASET
        )
        self.metadata_custom._add(
            ModelCustomMetadataItem(
                key=key,
                value=oci_storage_path,
                description=f"The path where {data_type} dataset path are stored on the object storage.",
                category=MetadataCustomCategory.TRAINING_AND_VALIDATION_DATASETS,
            ),
            replace=True,
        )

    def _save_data_shape(self, data, data_type):
        key = (
            MetadataCustomKeys.TRAINING_DATASET_SIZE
            if data_type == "training"
            else MetadataCustomKeys.VALIDATION_DATASET_SIZE
        )
        self.metadata_custom._add(
            ModelCustomMetadataItem(
                key=MetadataCustomKeys.TRAINING_DATASET_SIZE
                if data_type == "training"
                else MetadataCustomKeys.VALIDATION_DATASET_SIZE,
                value=str(data.shape),
                description=f"The size of the {data_type} dataset.",
                category=MetadataCustomCategory.TRAINING_AND_VALIDATION_DATASETS,
            ),
            replace=True,
        )

    def _save_file_size(self, file_path, data_type):
        self.metadata_custom._add(
            ModelCustomMetadataItem(
                key=MetadataCustomKeys.TRAINING_DATASET_SIZE
                if data_type == "training"
                else MetadataCustomKeys.VALIDATION_DATASET_SIZE,
                value=str(os.stat(file_path).st_size) + " bytes",
                description=f"The {data_type} dataset size in bytes.",
                category=MetadataCustomCategory.TRAINING_AND_VALIDATION_DATASETS,
            ),
            replace=True,
        )

    def populate_schema(
        self,
        data_sample: ADSData = None,
        X_sample: Union[list, tuple, pd.DataFrame, pd.Series, np.ndarray] = None,
        y_sample: Union[list, tuple, pd.DataFrame, pd.Series, np.ndarray] = None,
    ):
        """
        Populates input schema and output schema.
        If the schema exceeds the limit of 32kb, save as json files to the artifact dir.

        Parameters
        ----------
        data_sample: ADSData
            A sample of the data that will be used to generate intput_schema and schema_output.
        X_sample: Union[list, tuple, pd.Series, np.ndarray, pd.DataFrame, dask.dataframe.core.Series, dask.dataframe.core.DataFrame]
            A sample of input data that will be used to generate input schema.
        y_sample: Union[list, tuple, pd.Series, np.ndarray, pd.DataFrame, dask.dataframe.core.Series, dask.dataframe.core.DataFrame]
            A sample of output data that will be used to generate output schema.
        """
        if data_sample is not None:
            assert isinstance(
                data_sample, ADSData
            ), "`data_sample` expects data of ADSData type. \
            Pass in to `X_sample` and `y_sample` for other data types."
            X_sample = data_sample.X
            y_sample = data_sample.y
        self.schema_input = self._populate_schema(
            X_sample,
            schema_file_name=INPUT_SCHEMA_FILE_NAME,
        )
        self.schema_output = self._populate_schema(
            y_sample,
            schema_file_name=OUTPUT_SCHEMA_FILE_NAME,
        )

    def _populate_schema(self, data, schema_file_name):
        """
        Populate schema and if the schema exceeds the limit of 32kb, save as a json file to the artifact_dir.

        Parameters
        ----------
        data_sample: ADSData
            A sample of the data that will be used to generate intput_schema and schema_output.
        X_sample: Union[list, tuple, pd.Series, np.ndarray, pd.DataFrame, dask.dataframe.core.Series, dask.dataframe.core.DataFrame]
            A sample of input data that will be used to generate input schema.
        y_sample: Union[list, tuple, pd.Series, np.ndarray, pd.DataFrame, dask.dataframe.core.Series, dask.dataframe.core.DataFrame]
            A sample of output data that will be used to generate output schema.
        """
        if data is not None:
            data = utils.to_dataframe(data)
            schema = data.ads.model_schema()
            schema.to_json_file(os.path.join(self.artifact_dir, schema_file_name))
            if self._validate_schema_size(schema, schema_file_name):
                return schema
        return Schema()

    def _validate_schema_size(self, schema, schema_file_name):
        result = False
        try:
            result = schema.validate_size()
        except SchemaSizeTooLarge:
            logger.warn(
                f"The {schema_file_name.replace('.json', '')} is larger than "
                f"`{METADATA_SIZE_LIMIT}` bytes and cannot be stored as model catalog metadata."
                f"It will be saved to {self.artifact_dir}/{schema_file_name}."
            )

        return result

    def introspect(self) -> pd.DataFrame:
        """Runs model introspection.

        Returns
        -------
        pd.DataFrame
           The introspection result in a DataFrame format.
        """
        return self._introspect()


class VersionConflictWarning(object):
    def __init__(self, version_conflicts):
        self.version_conflicts = version_conflicts

    def __str__(self):
        msg = "WARNING: Version conflicts found:"
        if len(self.version_conflicts) > 0:
            for lib in self.version_conflicts:
                msg += "\nInstalled: %s, Required: %s" % (
                    lib,
                    self.version_conflicts[lib],
                )
        return msg


def pip_install(package, options="-U"):
    package = re.sub(r"<|>", "=", package.split(",")[0])
    for output in execute(["pip", "install", options, package]):
        print(output, end="")


def execute(cmd):
    popen = subprocess.Popen(cmd, stdout=subprocess.PIPE, universal_newlines=True)
    for stdout_line in iter(popen.stdout.readline, ""):
        yield stdout_line
    popen.stdout.close()
    return_code = popen.wait()
    if return_code:
        raise subprocess.CalledProcessError(return_code, cmd)


def fetch_manifest_from_conda_location(env_location: str):
    """
    Convenient method to fetch manifest file from an environment.

    :param env_location: Absolute path to the environment
    :type env_location: str
    """
    manifest_location = None
    for file in os.listdir(env_location):
        if file.endswith("_manifest.yaml"):
            manifest_location = f"{env_location}/{file}"
            break
    env = {}
    if not manifest_location:
        raise Exception(
            f"Could not locate manifest file in the provided environment: {env_location}. Dir Listing - "
            f"{os.listdir(env_location)}"
        )

    with open(manifest_location) as mlf:
        env = yaml.load(mlf, Loader=yaml.FullLoader)
    manifest = env["manifest"]
    return manifest



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2021 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import json
import oci

from ads.common import auth as authutil
from ads.common import oci_client


class OCILanguage(object):  # pragma: no cover
    def __init__(self, auth=None):
        auth = auth if auth else authutil.api_keys()
        self.ai_client = oci_client.OCIClientFactory(**auth).ai_language

    @property
    def ner(self):
        detect_language_entities_details = (
            oci.ai_language.models.DetectLanguageEntitiesDetails(text=self.string)
        )
        output = self.ai_client.detect_language_entities(
            detect_language_entities_details
        )
        return json.loads(str(output.data))["entities"]

    @property
    def lang(self):
        detect_dominant_language_details = (
            oci.ai_language.models.DetectDominantLanguageDetails(text=self.string)
        )
        output = self.ai_client.detect_dominant_language(
            detect_dominant_language_details
        )
        return json.loads(str(output.data))

    @property
    def key_phrases(self):
        detect_language_key_phrases_details = (
            oci.ai_language.models.DetectLanguageKeyPhrasesDetails(text=self.string)
        )
        output = self.ai_client.detect_language_key_phrases(
            detect_language_key_phrases_details
        )
        return json.loads(str(output.data))["key_phrases"]

    @property
    def sentiment(self):
        detect_language_sentiments_details = (
            oci.ai_language.models.DetectLanguageSentimentsDetails(text=self.string)
        )
        output = self.ai_client.detect_language_sentiments(
            detect_language_sentiments_details
        )
        return json.loads(str(output.data))["aspects"]



import re
from ads.common.model_info_extractor import ModelInfoExtractor
from ads.common.model_metadata import Framework


class SklearnExtractor(ModelInfoExtractor):
    """Class that extract model metadata from sklearn models.

    Attributes
    ----------
    model: object
        The model to extract metadata from.
    estimator: object
        The estimator to extract metadata from.

    Methods
    -------
    framework(self) -> str
        Returns the framework of the model.
    algorithm(self) -> object
        Returns the algorithm of the model.
    version(self) -> str
        Returns the version of framework of the model.
    hyperparameter(self) -> dict
        Returns the hyperparameter of the model.
    """

    def __init__(self, model, estimator):
        self.model = model
        self.estimator = estimator

    def framework(self):
        """Extracts the framework of the model.

        Returns
        ----------
        str:
           The framework of the model.
        """
        return Framework.SCIKIT_LEARN

    def algorithm(self):
        """Extracts the algorithm of the model.

        Returns
        ----------
        object:
           The algorithm of the model.
        """
        return self.estimator

    def version(self):
        """Extracts the framework version of the model.

        Returns
        ----------
        str:
           The framework version of the model.
        """
        import sklearn

        return sklearn.__version__

    def hyperparameter(self):
        """Extracts the hyperparameters of the model.

        Returns
        ----------
        dict:
           The hyperparameters of the model.
        """
        hp_dict = self.model.get_params()
        # handle sklearn pipeline case
        if "steps" in hp_dict:
            for i, (k, v) in enumerate(hp_dict["steps"]):
                hp_dict["steps"][i] = {k: re.sub("[()]", "", str(v))}
                hp_dict[k] = re.sub("[()]", "", str(v))
        # handle sklearn model selection case
        elif "param_grid" in hp_dict:
            hp_dict["estimator"] = str(hp_dict["estimator"])
            for k, v in hp_dict["param_grid"].items():
                hp_dict["param_grid"][k] = v.tolist()
            hp_dict.update(self.model.best_params_)
        return hp_dict



from ads.common.model_metadata import Framework
from ads.common.model_info_extractor import ModelInfoExtractor


class LightgbmExtractor(ModelInfoExtractor):
    """Class that extract model metadata from lightgbm models.

    Attributes
    ----------
    model: object
        The model to extract metadata from.
    estimator: object
        The estimator to extract metadata from.

    Methods
    -------
    framework(self) -> str
        Returns the framework of the model.
    algorithm(self) -> object
        Returns the algorithm of the model.
    version(self) -> str
        Returns the version of framework of the model.
    hyperparameter(self) -> dict
        Returns the hyperparameter of the model.
    """

    def __init__(self, model, estimator):
        self.model = model
        self.estimator = estimator

    def framework(self):
        """Extracts the framework of the model.

        Returns
        ----------
        str:
           The framework of the model.
        """
        return Framework.LIGHT_GBM

    def algorithm(self):
        """Extracts the algorithm of the model.

        Returns
        ----------
        object:
           The algorithm of the model.
        """
        return self.estimator

    def version(self):
        """Extracts the framework version of the model.

        Returns
        ----------
        str:
           The framework version of the model.
        """
        import lightgbm

        return lightgbm.__version__

    def hyperparameter(self):
        """Extracts the hyperparameters of the model.

        Returns
        ----------
        dict:
           The hyperparameters of the model.
        """
        return self.model.get_params()



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2021 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import ast
import base64
import json
import oci
import uuid

from oci.vault import VaultsClientCompositeOperations
from oci.vault.models import Base64SecretContentDetails
from oci.vault.models import CreateSecretDetails
from oci.vault.models import UpdateSecretDetails

from ads.common import oci_client, auth
from ads.config import NB_SESSION_COMPARTMENT_OCID


class Vault:
    def __init__(
        self,
        vault_id: str,
        key_id: str,
        compartment_id=None,
        secret_client_auth=None,
        vault_client_auth=None,
    ):
        self.id = vault_id
        self.key_id = key_id
        self.secret_client_auth = (
            secret_client_auth if secret_client_auth else auth.default_signer()
        )
        self.vault_client_auth = (
            vault_client_auth if vault_client_auth else auth.default_signer()
        )

        self.compartment_id = (
            NB_SESSION_COMPARTMENT_OCID if compartment_id is None else compartment_id
        )
        if not isinstance(self.compartment_id, str):
            raise ValueError("compartment_id needs to be a string.")

        if self.compartment_id is None:
            raise ValueError("compartment_id needs to be specified.")

        self.secret_client = oci_client.OCIClientFactory(
            **self.secret_client_auth
        ).secret
        self.vaults_client_composite = VaultsClientCompositeOperations(
            oci_client.OCIClientFactory(**self.vault_client_auth).vault
        )

    def create_secret(
        self, value: dict, secret_name: str = None, description: str = None
    ) -> str:
        """
        Saves value into vault as a secret.

        Parameters
        ----------
        value: dict
            The value to store as a secret.
        secret_name: str, optional
            The name of the secret.
        description: str, optional
            The description of the secret.

        Returns
        -------
        The secret ocid that correspond to the value saved as a secret into vault.
        """
        # Encode the secret.
        secret_content_details = self._encode_secret(value)

        # Bundle the secret and metadata about it.
        secrets_details = CreateSecretDetails(
            compartment_id=self.compartment_id,
            description=description
            if description is not None
            else "Data Science service secret",
            secret_content=secret_content_details,
            secret_name=secret_name
            if secret_name is not None
            else "DataScienceSecret_{}".format(str(uuid.uuid4())[-6:]),
            vault_id=self.id,
            key_id=self.key_id,
        )

        # Store secret and wait for the secret to become active.
        secret = self.vaults_client_composite.create_secret_and_wait_for_state(
            create_secret_details=secrets_details,
            wait_for_states=[oci.vault.models.Secret.LIFECYCLE_STATE_ACTIVE],
        ).data
        return secret.id

    def update_secret(self, secret_id: str, secret_content: dict) -> str:
        """
        Updates content of a secret.

        Parameters
        ----------
        secret_id: str
            The secret id where the stored secret will be updated.
        secret_content: dict,
            The updated content.

        Returns
        -------
        The secret ocid with updated content.
        """

        # Encode the secret.
        secret_content_details = self._encode_secret(secret_content)

        # Store the details to update.
        secrets_details = UpdateSecretDetails(secret_content=secret_content_details)

        # Create new secret version and wait for the new version to become active.
        secret_update = self.vaults_client_composite.update_secret_and_wait_for_state(
            secret_id,
            secrets_details,
            wait_for_states=[oci.vault.models.Secret.LIFECYCLE_STATE_ACTIVE],
        ).data

        return secret_update.id

    def get_secret(self, secret_id: str) -> dict:
        """
        Retrieve secret content based on the secret ocid provided

        Parameters
        ----------
        secret_id: str
            The secret ocid.

        Returns
        -------
        The secret content as a dictionary.
        """
        secret_bundle = self.secret_client.get_secret_bundle(secret_id)
        secret_content = self._secret_to_dict(
            secret_bundle.data.secret_bundle_content.content
        )
        return ast.literal_eval(secret_content)

    def _encode_secret(self, secret_content):
        secret_content_details = Base64SecretContentDetails(
            content_type=oci.vault.models.SecretContentDetails.CONTENT_TYPE_BASE64,
            stage=oci.vault.models.SecretContentDetails.STAGE_CURRENT,
            content=self._dict_to_secret(secret_content),
        )
        return secret_content_details

    @staticmethod
    def _dict_to_secret(values):
        return base64.b64encode(json.dumps(values).encode("ascii")).decode("ascii")

    @staticmethod
    def _secret_to_dict(secret_content):
        return base64.b64decode(secret_content.encode("ascii")).decode("ascii")



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import numpy as np
from scipy.stats import skew
from sklearn import preprocessing

from ads.common import utils
from ads.type_discovery.typed_feature import CategoricalTypedFeature
from ads.type_discovery.typed_feature import ContinuousTypedFeature
from ads.type_discovery.typed_feature import OrdinalTypedFeature


class TargetVariable:
    """
    This class provides target specific APIs.
    """

    def __init__(self, sampled_ds, target, target_type):
        self.sampled_ds = sampled_ds
        self.name = target
        self.type = target_type
        self.target_vals = None
        if isinstance(self.type, CategoricalTypedFeature):
            self.target_vals = self.sampled_ds.sampled_df[target].unique().tolist()
            self.skewness = np.abs(
                skew(
                    preprocessing.LabelEncoder().fit_transform(
                        self.sampled_ds.sampled_df[target]
                    )
                )
            )
        elif isinstance(self.type, ContinuousTypedFeature):
            try:
                self.skewness = np.abs(self.sampled_ds.sampled_df[target].skew())
            except TypeError as e:
                self.skewness = np.abs(
                    self.sampled_ds.sampled_df[target].astype("float").skew()
                )
        else:
            # can also be DateTimeTypedFeature IPAddressTypedFeature PhoneNumberTypedFeature GISTypedFeature
            # AddressTypedFeature DocumentTypedFeature ZipcodeTypedFeature UnknownTypedFeature ConstantTypedFeature
            # CreditCardTypedFeature OrdinalTypedFeature
            self.skewness = None
        self.numeric_columns = (
            self.sampled_ds.sampled_df._get_numeric_data().columns.values
        )

    def show_in_notebook(self, feature_names=None):  # pragma: no cover
        """
        Plot target distribution or target versus feature relation.

        Parameters
        ----------
        feature_names: list, Optional
            Plot target against a list of features.
            Display target distribution if feature_names is not provided.
        """
        if not utils.is_notebook():
            print("show_in_notebook called but not in notebook environment")
            return

        verbose = True
        if feature_names is not None:
            for feature_name in feature_names:
                self.sampled_ds.plot(
                    feature_name, self.name, verbose=verbose
                ).show_in_notebook()
                verbose = False
        else:
            self.sampled_ds.plot(self.name, verbose=False).show_in_notebook()

    def is_balanced(self):
        """
        Returns True if the target is balanced, False otherwise.

        Returns
        -------
        is_balanced: bool
        """
        if isinstance(self.type, CategoricalTypedFeature) or isinstance(
            self.type, ContinuousTypedFeature
        ):
            return self.skewness < 0.2
        else:
            return True



"""Contains Mixins for integrating OCI data models
"""
import inspect
import json
import logging
import os
import re
from datetime import date, datetime
from typing import Callable, Optional, Union

import oci
from dateutil import tz
from dateutil.parser import parse
from oci._vendor import six

from ads.common.auth import default_signer
from ads.common.oci_client import OCIClientFactory

logger = logging.getLogger(__name__)


class OCIClientMixin:
    """Mixin class for representing OCI resource/service with OCI client.

    Most OCI requests requires a client for making the connection.
    Usually the same client will be used for the requests related the same resource/service type.
    This Mixin adds a client property to simplify accessing the client.
    The actual client will be initialize lazily so that it is not required for a sub-class
    To use the client, sub-class should override the init_client() method pass in the "client" keyword argument.
    For example:

    @staticmethod
    def init_client(**kwargs) -> oci.logging.LoggingManagementClient:
        return cls._init_client(client=oci.logging.LoggingManagementClient, **kwargs)

    Authentication
    --------------
    The init_client() method is a class method. The call to this method may come from an instance or a class.
    When the method is called,
        the client will be initialized with the credentials passed in by the "auth" keyword argument.
    In order to allow the user to specify "auth", all methods using init_client() should expose "auth" as an argument.
    If "auth" is not specified,
        The default auth configured at ADS level with ads.set_auth() will be used.
    """

    @classmethod
    def init_client(cls, **kwargs):
        """Initializes the OCI client specified in the "client" keyword argument
        Sub-class should override this method and call cls._init_client(client=OCI_CLIENT)
        """
        return cls._init_client(**kwargs)

    @classmethod
    def _init_client(cls, auth: dict = None, **kwargs):
        """Initializes the OCI client specified in the "client" keyword argument

        Parameters
        ----------
        auth : dict
            ADS auth dictionary for OCI authentication.
            This can be generated by calling ads.common.auth.api_keys() or ads.common.auth.resource_principal()
            If this is None, ads.common.default_signer(client_kwargs) will be used.
        **kwargs :
            client is required for specifying the class of OCI client
            Additional keyword arguments will be passed to ads.common.default_signer() when auth is None.

        Returns
        -------
            An instance of OCI client.

        """
        client_args = dict(retry_strategy=oci.retry.DEFAULT_RETRY_STRATEGY)
        client_class = kwargs.pop("client", None)
        if not client_class:
            raise ValueError("client is not specified in keyword argument.")

        client_args.update(kwargs)

        # If auth is empty at this time,
        # it mean the class authentication is not configured at the time when the instance was created, or
        # the call is from a class without auth configured.
        # The ADS default will be used.
        if not auth:
            auth = default_signer(client_args)

        return OCIClientFactory(
            **auth,
        ).create_client(client_class)

    def __init__(self) -> None:
        """Initializes a service/resource with OCI client as a property."""
        super().__init__()
        self._client = None

    @property
    def client(self):
        """OCI client"""
        if self._client is None:
            self._client = self.init_client()
        return self._client


class OCISerializableMixin(OCIClientMixin):
    """Mixin class containing OCI serialization/de-serialization methods.
    These methods are copied and modified from the OCI BaseClient.

    """

    type_mappings = None

    def serialize(self):
        """Serialize the model to a dictionary that is ready to be send to OCI API.

        Returns
        -------
        dict
            A dictionary that is ready to be send to OCI API.

        """
        return self.client.base_client.sanitize_for_serialization(self)

    @staticmethod
    def _parse_kwargs(attribute_map: dict, **kwargs):
        """Parse kwargs to make all key in camel format."""
        parsed_kwargs = {}
        for key, val in kwargs.items():
            if key in attribute_map:
                parsed_kwargs[attribute_map[key]] = val
            else:
                parsed_kwargs[key] = val

        return parsed_kwargs

    @classmethod
    def deserialize(cls, data, to_cls):
        """De-serialize data from dictionary to an OCI model"""
        if cls.type_mappings is None:
            cls.type_mappings = cls.init_client().base_client.type_mappings

        if data is None:
            return None

        # This is a work around for enums not being present
        # in the type mappings.
        # See OraclePythonSdkCodegen removeEnumsFromModelGeneration().
        if to_cls in cls.type_mappings:
            to_cls = cls.type_mappings[to_cls]

            if isinstance(data, dict):
                data = cls._parse_kwargs(to_cls().attribute_map, **data)
        else:
            return cls.__deserialize_primitive(data, to_cls)

        if hasattr(to_cls, "get_subtype"):
            # Return the object as is if it is already deserialized.
            if isinstance(data, to_cls) or issubclass(data.__class__, to_cls):
                return data

            # Use the discriminator value to get the correct subtype.
            to_cls = to_cls.get_subtype(data)  # get_subtype returns a str
            to_cls = cls.type_mappings[to_cls]

            # kwargs needs to be parsed again as there are more attributes in the sub types.
            if isinstance(data, dict):
                data = cls._parse_kwargs(to_cls().attribute_map, **data)

        if to_cls in [int, float, six.u, bool]:
            return cls.__deserialize_primitive(data, to_cls)
        elif to_cls == object:
            return data
        elif to_cls == date:
            return cls.__deserialize_date(data)
        elif to_cls == datetime:
            return cls.__deserialize_datetime(data)
        else:
            return cls.__deserialize_model(data, to_cls)

    @classmethod
    def __deserialize_model(cls, data, to_cls):
        """De-serializes list or dict to model."""
        if isinstance(data, to_cls):
            return data

        instance = to_cls()

        for attr, attr_type in instance.swagger_types.items():
            property = instance.attribute_map[attr]
            if property in data:
                value = data[property]
                setattr(instance, attr, cls.deserialize(value, attr_type))

        return instance

    @staticmethod
    def __deserialize_primitive(data, cls):
        """De-serializes string to primitive type."""
        try:
            value = cls(data)
        except UnicodeEncodeError:
            value = six.u(data)
        except TypeError:
            value = data
        return value

    @staticmethod
    def __deserialize_date(string):
        """De-serializes string to date."""
        try:
            return parse(string).date()
        except ImportError:
            return string
        except ValueError:
            raise Exception("Failed to parse `{0}` into a date object".format(string))

    @staticmethod
    def __deserialize_datetime(string):
        """De-serializes string to datetime.

        The string should be in iso8601 datetime format.
        """
        if isinstance(string, datetime):
            return string
        try:
            # If this parser creates a date without raising an exception
            # then the time zone is utc and needs to be set.
            naivedatetime = datetime.strptime(string, "%Y-%m-%dT%H:%M:%S.%fZ")
            awaredatetime = naivedatetime.replace(tzinfo=tz.tzutc())
            return awaredatetime

        except ValueError:
            try:
                return parse(string)
            except ImportError:
                return string
            except ValueError:
                raise Exception(
                    "Failed to parse `{0}` into a datetime object".format(string)
                )
        except ImportError:
            return string


class OCIModelMixin(OCISerializableMixin):
    """Mixin class to operate OCI model.
    OCI resources are represented by models in the OCI Python SDK.

    Unifying OCI models for the same resource
    -----------------------------------------
    OCI SDK uses different models to represent the same resource for different operations.
    For example, CreateLogDetails is used when creating a log resource,
        while LogSummary is returned when listing the log resources.
    However, both CreateLogDetails and LogSummary have the same commonly used attribute like
        compartment_id, display_name, log_type, etc.
    In general, there is a class with a super set of all properties.
    For example, the Log class contains all properties of CreateLogDetails and LogSummary,
        as well as other classes representing an OCI log resource.
    A subclass can be implemented with this Mixin to unify the OCI models,
        so that all properties are available to the user.
    For example, if we define the Mixin model as ``class OCILog(OCIModelMixin, oci.logging.models.Log)``,
        users will be able to access properties like ``OCILog().display_name``
    Since this sub-class contains all the properties, it can be converted to any related OCI model in an operation.
    For example, we can create ``CreateLogDetails`` from ``OCILog`` by extracting a subset of the properties.
    When listing the resources, properties from ``LogSummary`` can be used to update
        the corresponding properties of ``OCILog``.
    Such convertion can be done be the generic methods provided by this Mixin.
    Although OCI SDK accepts dictionary (JSON) data instead of objects like CreateLogDetails when creating or
        updating the resource, the structure and keys of the dictionary is not easy for a user to remember.
    It is also unnecessary for the users to construct the entire dictionary if they only want to update a single value.

    This Mixin class should be the first parent as the class from OCI SDK does not call ``super().__init__()``
        in its ``__init__()`` constructor.
    Mixin properties may not be intialized correctly if ``super().__init__()`` is not called.


    Provide generic methods for CRUDL operations
    --------------------------------------------
    Since OCI SDK uses different models in CRUDL operations,
        this Mixin provides the following method to convert between them.
    An OCI model instance can be any OCI model of the resource containing some properties, e.g. LogSummary
    ``from_oci_model()`` static method can be used to initialize a new instance from an OCI model instance.
    ``update_from_oci_model()`` can be used to update the existing properties from an OCI model instance.
    ``to_oci_model()`` can be used to extract properties from the Mixin model to OCI model.

    """

    # Regex pattern matching the module name of an OCI model.
    OCI_MODEL_PATTERN = r"oci.[^.]+\.models[\..*]?"
    OCI_LIST_METHOD = None

    CONS_COMPARTMENT_ID = "compartment_id"

    @staticmethod
    def check_compartment_id(compartment_id: Optional[str]) -> str:
        """Checks if a compartment ID has value and
            return the value from NB_SESSION_COMPARTMENT_OCID environment variable if it is not specified.

        Parameters
        ----------
        compartment_id : str
            Compartment OCID or None

        Returns
        -------
        type
            str: Compartment OCID

        Raises
        ------
        ValueError
            compartment_id is not specified and NB_SESSION_COMPARTMENT_OCID environment variable is not set

        """
        compartment_id = (
            compartment_id
            if compartment_id
            else os.environ.get("NB_SESSION_COMPARTMENT_OCID")
        )
        if not compartment_id:
            raise ValueError("Please specify compartment OCID.")
        return compartment_id

    @staticmethod
    def camel_to_snake(s: str) -> str:
        """Converts camel style string to snake style.

        Parameters
        ----------
        s : str
            The string to be converted


        Returns
        -------
        str
            A string in snake format.

        """
        return re.sub(r"(?<!^)(?=[A-Z])", "_", s).lower()

    @classmethod
    def list_resource(cls, compartment_id: str = None, **kwargs) -> list:
        """Generic method to list OCI resources

        Parameters
        ----------
        compartment_id : str
            Compartment ID of the OCI resources. Defaults to None.
            If compartment_id is not specified,
            the value of NB_SESSION_COMPARTMENT_OCID in environment variable will be used.
        **kwargs :
            Additional keyword arguments to filter the resource.
            The kwargs are passed into OCI API.

        Returns
        -------
        list
            A list of OCI resources

        Raises
        ------
        NotImplementedError
            List method is not supported or implemented.

        """
        items = oci.pagination.list_call_get_all_results(
            cls._find_oci_method("list"),
            cls.check_compartment_id(compartment_id),
            **kwargs,
        ).data
        return [cls.from_oci_model(item) for item in items]

    @classmethod
    def _find_oci_parent(cls):
        """Finds the parent OCI model class

        Parameters
        ----------

        Returns
        -------
        class
            An OCI model class

        Raises
        ------
        AttributeError
            If the class if not inherit from an OCI model.

        """
        oci_model = None
        for parent in inspect.getmro(cls):
            if re.match(OCIModelMixin.OCI_MODEL_PATTERN, parent.__module__):
                oci_model = parent
                break
        if not oci_model:
            raise AttributeError(f"{cls.__name__} is not inherited from an OCI model.")
        return oci_model

    @classmethod
    def _find_oci_method(cls, method: str) -> Callable:
        """Finds the OCI method for operations like get/list

        Parameters
        ----------
        method : str
            The operation to be performed, e.g. get, list, delete


        Returns
        -------
        callable
            The method from OCI client to perform the operation.

        """
        client = cls.init_client()

        user_define_method = f"oci_{method}_method".upper()
        if hasattr(cls, user_define_method) and getattr(cls, user_define_method):
            return getattr(cls, user_define_method).__get__(client, client.__class__)

        oci_class = cls._find_oci_parent()
        if method == "list":
            method_name = f"{method}_{cls.camel_to_snake(oci_class.__name__)}s"
        else:
            method_name = f"{method}_{cls.camel_to_snake(oci_class.__name__)}"

        if hasattr(client, method_name):
            return getattr(client, method_name)

        raise NotImplementedError(
            f"{str(oci_class)} does not have {method_name} method. "
            f"Please define {user_define_method} in {str(cls)}"
        )

    @classmethod
    def from_oci_model(cls, oci_instance):
        """Initialize an instance from an instance of OCI model.

        Parameters
        ----------
        oci_instance :
            An instance of an OCI model.

        """
        oci_model = cls._find_oci_parent()
        kwargs = {}
        for attr in oci_model().swagger_types.keys():
            if hasattr(oci_instance, attr):
                kwargs[attr] = getattr(oci_instance, attr)
        return cls(**kwargs)

    @classmethod
    def from_dict(cls, data):
        """Initialize an instance from a dictionary.

        Parameters
        ----------
        data : dict
            A dictionary containing the properties to initialize the class.

        """
        instance = cls()
        oci_instance = cls.deserialize(data)
        instance.update_from_oci_model(oci_instance)
        return instance

    @classmethod
    def deserialize(cls, data, to_cls=None):
        """Deserialize data

        Parameters
        ----------
        data :

        to_cls :
             (Default value = None)

        Returns
        -------

        """
        if to_cls is None:
            oci_model = cls._find_oci_parent()
            to_cls = oci_model.__name__
        return super().deserialize(data, to_cls)

    @classmethod
    def _get(cls, ocid):
        get_method = cls._find_oci_method("get")
        return get_method(ocid).data

    @classmethod
    def from_ocid(cls, ocid: str):
        """Initializes an object from OCID

        Parameters
        ----------
        ocid : str
            The OCID of the object

        """
        oci_instance = cls._get(ocid)
        return cls.from_oci_model(oci_instance)

    def __init__(self, auth: dict = None, **kwargs) -> None:
        # Initialize an empty object
        super().__init__()
        if auth:
            self._client_auth.update(auth)
        # Load properties from environment
        self.load_properties_from_env()
        if kwargs:
            self.update_from_oci_model(self.deserialize(kwargs))
        # Retrieve the compartment ID if user specified the OCID of the resource but not compartment ID
        if (
            self.id
            and hasattr(self, self.CONS_COMPARTMENT_ID)
            and not getattr(self, self.CONS_COMPARTMENT_ID)
        ):
            from .oci_resource import OCIResource

            setattr(
                self, self.CONS_COMPARTMENT_ID, OCIResource.get_compartment_id(self.id)
            )

    @property
    def name(self):
        """Display name of the object"""
        if hasattr(self, "display_name"):
            return getattr(self, "display_name")
        raise NotImplementedError()

    def load_properties_from_env(self):
        """Loads properties from the environment"""
        env_var_mapping = {
            "project_id": "PROJECT_OCID",
            OCIModelMixin.CONS_COMPARTMENT_ID: "NB_SESSION_COMPARTMENT_OCID",
        }
        for attr, env_name in env_var_mapping.items():
            if hasattr(self, attr) and not getattr(self, attr):
                setattr(self, attr, os.environ.get(env_name))

    def to_oci_model(self, oci_model):
        """Converts the object into an instance of OCI data model.

        Parameters
        ----------
        oci_model : class or str
            The OCI model to be converted to. This can be a string of the model name.
        type_mapping : dict
            A dictionary mapping the models.
            Returns: An instance of the oci_model

        Returns
        -------

        """
        if isinstance(oci_model, str):
            oci_model_name = oci_model
        else:
            oci_model_name = oci_model.__name__
        data = json.dumps(self.to_dict()).encode("utf8")
        return self.client.base_client.deserialize_response_data(data, oci_model_name)

    @staticmethod
    def flatten(data: dict) -> dict:
        """Flattens a nested dictionary.

        Parameters
        ----------
        data : A nested dictionary


        Returns
        -------
        dict
            The flattened dictionary.

        """
        flatten_dict = {}
        for key, value in data.items():
            if isinstance(value, dict):
                flatten_dict.update(OCIModelMixin.flatten(value))
            else:
                flatten_dict[key] = value
        return flatten_dict

    def to_dict(self, flatten: bool = False) -> dict:
        """Converts the properties to a dictionary

        Parameters
        ----------
        flatten :
             (Default value = False)

        Returns
        -------

        """
        data = self.serialize()
        if flatten:
            data = OCIModelMixin.flatten(data)
        return data

    def update_from_oci_model(self, oci_model_instance):
        """Updates the properties from OCI model with the same properties.

        Parameters
        ----------
        oci_model_instance :
            An instance of OCI model, which should have the same properties of this class.

        """
        for attr in self.swagger_types.keys():
            if hasattr(oci_model_instance, attr) and getattr(oci_model_instance, attr):
                setattr(self, attr, getattr(oci_model_instance, attr))
        return self

    def sync(self):
        """Refreshes the properties of the object from OCI"""
        return self.update_from_oci_model(self.from_ocid(self.id))

    def delete(self):
        """Deletes the resource"""
        delete_method = self._find_oci_method("delete")
        delete_method(self.id)
        return self


class OCIWorkRequestMixin:
    """Mixin class containing methods related to OCI work request"""

    def wait_for_work_request(
        self,
        work_request_id: str,
        wait_for_state: Union[str, tuple],
        max_wait_seconds: int = None,
        wait_interval_seconds: int = None,
    ):
        """Wait for a work request to be completed.

        Parameters
        ----------
        work_request_id : str
            OCI work request ID
        wait_for_state : str or tuple
            The state to wait for. Must be a tuple for multiple states.
        max_wait_seconds : int
            Max wait seconds for the work request. Defaults to None (Default value from OCI SDK will be used).
        wait_interval_seconds : int
            Interval in seconds between each status check. Defaults to None (Default value from OCI SDK will be used).

        Returns
        -------
        Response
            OCI API Response

        """
        result = None
        if hasattr(self.client, "get_work_request") and callable(
            getattr(self.client, "get_work_request")
        ):
            try:
                wait_period_kwargs = {}
                if max_wait_seconds is not None:
                    wait_period_kwargs["max_wait_seconds"] = max_wait_seconds
                if wait_interval_seconds is not None:
                    wait_period_kwargs["max_interval_seconds"] = wait_interval_seconds

                logger.debug(
                    "Action completed. Waiting until the work request has entered state: %s",
                    wait_for_state,
                )
                result = oci.wait_until(
                    self.client,
                    self.client.get_work_request(work_request_id),
                    "status",
                    wait_for_state,
                    **wait_period_kwargs,
                )
            except oci.exceptions.MaximumWaitTimeExceeded:
                # If we fail, we should show an error, but we should still provide the information to the customer
                logger.error(
                    "Failed to wait until the work request %s entered the specified state. Maximum wait time exceeded",
                    work_request_id,
                )
            except Exception:
                logger.error(
                    "Encountered error while waiting for work request to enter the specified state."
                )
                raise
        else:
            logger.error(
                "%s does not support wait for the work request to enter the specified state",
                str(self.client.__class__),
            )
        return result



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2021 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import logging
import os
import tempfile
import zipfile
from time import time
from typing import Dict, Optional, List, Union, Iterator

import cx_Oracle
import pandas as pd
import numpy as np

logger = logging.getLogger("ads.oracle_connector")

# logging.basicConfig(
#    format="[%(asctime)s] (%(levelname)s): %(message)s", level=logging.INFO
# )


class OracleRDBMSConnection(cx_Oracle.Connection):
    def __init__(self, *args, wallet_file=None, **kwargs):
        self.temp_dir = None
        self.tns_entries = {}
        if wallet_file:
            self._setup_wallet_file(wallet_file)
            if len(args) > 2:
                dsn = args[2]
            if dsn in self.tns_entries:
                dsn = self.tns_entries[dsn]
                if len(args) > 2:
                    args = args[:2] + (dsn,) + args[3:]
                else:
                    kwargs = kwargs.copy()
                    kwargs["dsn"] = dsn

        super().__init__(*args, **kwargs)

        logger.info(
            f"cx_Oracle version: {'.'.join([str(x) for x in cx_Oracle.clientversion()[:2]])}"
        )
        logger.info(
            f"RDBMS version: {'.'.join([str(x) for x in self.version.split('.')[:2]])}"
        )

    def __del__(self):
        if self.temp_dir is not None:
            self.temp_dir.cleanup()

    def _setup_wallet_file(self, wallet_file: str):
        # extract files in wallet zip to a temporary directory
        self.temp_dir = tempfile.TemporaryDirectory()
        zipfile.ZipFile(wallet_file).extractall(self.temp_dir.name)

        # parse tnsnames.ora to get list of entries and modify them to include the wallet location
        fname = os.path.join(self.temp_dir.name, "tnsnames.ora")
        for line in open(fname):
            pos = line.find(" = ")
            if pos < 0:
                continue
            name = line[:pos]
            entry = line[pos + 3 :].strip()
            key_phrase = "(security="
            pos = entry.find(key_phrase) + len(key_phrase)
            wallet_entry = f"(MY_WALLET_DIRECTORY={self.temp_dir.name})"
            entry = entry[:pos] + wallet_entry + entry[pos:]
            self.tns_entries[name] = entry

    def insert(
        self, table_name: str, df: pd.DataFrame, if_exists: str, batch_size=100000
    ):

        if if_exists not in ["fail", "replace", "append"]:
            raise ValueError(
                "Unknown option `if_exists`={if_exists}. Valid options are 'fail', 'replace', 'append'"
            )
        start_time = time()

        df_orcl = df.copy()
        # "object" type can be actual objects, or just plain strings, when inserting into the
        # database we need to stringify these so they can be represented in a VARCHAR2 column

        # object_columns = df_orcl.select_dtypes(include=["object"]).columns
        # df_orcl = df_orcl.where(pd.notnull(df_orcl), None)
        # df_orcl[object_columns] = df_orcl[object_columns].astype(str)

        # prep column names for valid Oracle column names (alpha + # $ _)
        df_orcl.columns = df_orcl.columns.str.replace(r"\W+", "_", regex=True)
        table_exist = True
        with self.cursor() as cursor:

            if if_exists != "replace":
                try:
                    cursor.execute(f"SELECT 1 from {table_name} FETCH NEXT 1 ROWS ONLY")
                except Exception as e:
                    table_exist = False
                if if_exists == "fail" and table_exist:
                    raise ValueError(
                        f"Table {table_name} already exists. Set `if_exists`='replace' or 'append' to replace or append to the existing table"
                    )
            # Oracle doesn't have boolean so convert to 1/0
            df_orcl = df_orcl.where(
                df.applymap(type) != bool, df_orcl.replace({True: 1, False: 0})
            )

            type_mappings = {
                "bool": "NUMBER(1)",
                "int16": "INTEGER",
                "int32": "INTEGER",
                "int64": "INTEGER",
                "float16": "FLOAT",
                "float32": "FLOAT",
                "float64": "FLOAT",
                "datetime64": "TIMESTAMP",
            }
            # add in any string types as Oracle's VARCHAR type setting length to accomodate longest
            def get_max_str_len(df, column):
                max_str_len = df[column].dropna().astype(str).str.len().max()
                if max_str_len >= 4000:
                    raise Exception(
                        f'Error: Column "{column}" is too long to store in a VARCHAR2 column'
                    )
                return max_str_len

            longest_string_column = max(
                (
                    get_max_str_len(df_orcl, c)
                    for c in df_orcl.select_dtypes(
                        include=["object", "category"]
                    ).columns
                ),
                default=0,
            )

            logger.debug(f"Max string column value: {longest_string_column}")

            datatypes = {
                c: f"VARCHAR2({get_max_str_len(df_orcl, c)})"
                for c in df_orcl.select_dtypes(include=["object", "category"]).columns
            }

            for df_type, orcl_type in type_mappings.items():

                datatypes.update(
                    {
                        column: orcl_type
                        for column in df_orcl.select_dtypes(include=df_type).columns
                    }
                )

            if set(datatypes.keys()) != set(df_orcl.columns):
                raise Exception(
                    f"unable to determine oracle data type to use for column(s): {', '.join(set(df_orcl.columns)-set(datatypes.keys()))}"
                )

            # create table
            if if_exists == "replace":
                try:
                    cursor.execute(f"drop table {table_name}")
                except:
                    logger.info(f"Table {table_name} does not exist")
            if if_exists == "replace" or not table_exist:
                sql = (
                    f"create table {table_name} ("
                    + ", ".join([f"{col} {datatypes[col]}" for col in df_orcl.columns])
                    + ")"
                )
                logger.info(sql)
                cursor.execute(sql)

            # insert
            bind_columns = ", ".join([f"{col}" for col in df_orcl.columns])
            bind_variables = ", ".join([f":{col}" for col in df_orcl.columns])
            sql = f"insert into {table_name}({bind_columns}) values({bind_variables})"

            logger.info(sql)

            # prevent buffer reallocations by locking in the longest string value
            # cursor.setinputsizes(None, longest_string_column)

            # replace NaN with None before turning into database records, important - don't
            # do this earlier in the logic because it can change the pandas
            # data types

            record_data = list(df_orcl.replace({np.nan: None}).itertuples(index=False))

            def chunks(lst: List, batch_size: int):
                """Yield successive batch_size chunks from lst."""
                for i in range(0, len(lst), batch_size):
                    yield lst[i : i + batch_size]

            for batch in chunks(record_data, batch_size=batch_size):

                cursor.executemany(sql, batch, batcherrors=True)

                for error in cursor.getbatcherrors():
                    logger.error(
                        f"Error: {error.message}, at row offset: {error.offset}"
                    )
                    raise RuntimeError(
                        f"Error: {error.message}, at row offset: {error.offset}"
                    )

            self.commit()

            duration = time() - start_time
            logger.info(
                f"inserted {df_orcl.shape[0]} rows at {df_orcl.shape[0]/duration:.2f} rows/seconds"
            )

    def _fetch_by_batch(self, cursor, chunksize):
        while True:
            rows = cursor.fetchmany(chunksize)
            if rows:
                yield rows
            else:
                break

    def query(
        self, sql: str, bind_variables: Optional[Dict], chunksize=None
    ) -> Union[pd.DataFrame, Iterator[pd.DataFrame]]:

        start_time = time()

        cursor = self.cursor()
        cursor.arraysize = 50000

        if chunksize:
            logger.info(f"Chunksize is {chunksize}")
            cursor.execute(sql, **bind_variables)
            columns = [row[0] for row in cursor.description]
            df = iter(
                (
                    pd.DataFrame(data=rows, columns=columns)
                    for rows in self._fetch_by_batch(cursor, chunksize)
                )
            )

        else:
            df = pd.DataFrame(
                cursor.execute(sql, **bind_variables),
                columns=[row[0] for row in cursor.description],
            )
            duration = time() - start_time
            logger.info(
                f"fetched {df.shape[0]} rows at {df.shape[0]/duration:.2f} rows/seconds"
            )

        return df


class OracleDBAccessMixin:
    @staticmethod
    def read_sql(
        sql: str,
        connection_parameters: dict,
        bind_variables: Dict = {},
        chunksize: Optional[int] = None,
    ):
        """Read SQL query from oracle database into a DataFrame.

        Parameters
        ----------
        sql: str
            SQL query to be executed.
        connection_parameters: dict
            A dictionary of connection_parameters - {"user_name":"", "password":"", "service_name":"", "wallet_location":""}
        bind_variables: Optional[Dict]
            Key value of pair of bind variables and corresponding values
        chunksize: Optional[int], default None
            If specified, return an iterator where `chunksize` is the number of rows to include in each chunk.

        Returns:
        -------
            DataFrame or Iterator[DataFrame]
                DataFrame or Iterator[DataFrame].

        Examples
        --------
        >>> connection_parameters = {
                "user_name": "<username>",
                "password": "<password>",
                "service_name": "{service_name}_{high|med|low}",
                "wallet_location": "/full/path/to/my_wallet.zip",
            }
        >>> import pandas as pd
        >>> import ads
        >>> df = pd.DataFrame.ads.read_sql("SELECT * from Employee", connection_parameters=connection_parameters)
        >>> df_with_bind = pd.DataFrame.ads.read_sql("SELECT * from EMPLOYEE WHERE EMPLOYEE_ID = :ID", bind_variables={"ID":"121212", connection_parameters=connection_parameters)


        """

        return OracleRDBMSConnection(
            connection_parameters["user_name"],
            connection_parameters["password"],
            connection_parameters["service_name"],
            wallet_file=connection_parameters["wallet_location"],
        ).query(sql, bind_variables=bind_variables, chunksize=chunksize)

    def to_sql(
        self,
        table_name: str,
        connection_parameters: dict,
        if_exists: str = "fail",
        batch_size=100000,
    ):
        """.

        Parameters
        ----------
        table_name: str
            Name of SQL table.
        connection_parameters: dict
            A dictionary of connection_parameters - {"user_name":"", "password":"", "service_name":"", "wallet_location":""}
        if_exists: : {'fail', 'replace', 'append'}, default 'fail'
            How to behave if the table already exists.
            * fail: Raise a ValueError. If table exists, do nothing
            * replace: Drop the table before inserting new values. If table exists, drop it, recreate it, and insert data.
            * append: Insert new values to the existing table. If table exists, insert data. Create if does not exist.
        batch_size: int, default 100000
            Inserting in batches improves insertion performance. Choose this value based on available memore and network bandwidth.


        Returns:
        -------
            None
                Nothing.
        Examples
        --------
        >>> connection_parameters = {
                "user_name": "<username>",
                "password": "<password>",
                "service_name": "{service_name}_{high|med|low}",
                "wallet_location": "/full/path/to/my_wallet.zip",
            }
        >>> import pandas as pd
        >>> import ads
        >>> df2 = pd.read_csv("my/data/csv")
        >>> df2.ads.to_sql("MY_DATA_CSV", connection_parameters=connection_parameters)
        """
        if if_exists not in ["fail", "replace", "append"]:
            raise ValueError(
                f"Unknown option `if_exists`={if_exists}. Valid options are 'fail', 'replace', 'append'"
            )
        return OracleRDBMSConnection(
            connection_parameters["user_name"],
            connection_parameters["password"],
            connection_parameters["service_name"],
            wallet_file=connection_parameters["wallet_location"],
        ).insert(table_name, self._obj, if_exists)



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import sys
import os
import json
from setuptools import setup, find_packages

install_requires = [
    "bokeh>=2.3.0",
    "cerberus>=1.3.4",
    "cloudpickle>=1.6.0",
    "cx-Oracle>=8.0",
    "dask-ml==1.8.0",
    "dask[array]==2.30.0",
    "dask[complete]==2.30.0",
    "datefinder>=0.7.1",
    "distributed==2.30.0",
    "fastavro==0.24.2",
    "fastparquet==0.7.1",
    "fdk>=0.1.18",
    "folium>=0.12.1",
    "fsspec>=0.8.7",
    "geopandas>=0.9.0",
    "gitpython>=3.1.2",
    "graphviz<0.17",
    "htmllistparse>=0.6.0",
    "ipython>=7.23.1,< 8.0",
    "ipywidgets==7.6.3",
    "jinja2>=2.11.2",
    "matplotlib>=3.1.3",
    "numexpr>=2.7.3",
    "numpy>=1.20.1",
    "oci>=2.43.1",
    "ocifs>=0.1.5",
    "onnx==1.7.0",
    "onnxmltools==1.7.0",
    "onnxruntime==1.4.0",
    "openpyxl==3.0.7",
    "optuna==2.9.0",
    "pandas>1.2.1,<1.4",
    "pandavro==1.6.0",
    "py-cpuinfo>=7.0.0",
    "pyarrow>=3.0,<4.0",
    "python-snappy==0.6.0",
    "python_jsonschema_objects>=0.3.13",
    "scikit-learn==0.23.2",
    "scipy==1.5.4",
    "seaborn>=0.11.0",
    "six>=1.14.0",
    "sqlalchemy>=1.4.1",
    "tabulate>=0.8.9",
    "tqdm>=4.59.0",
    "wordcloud>=1.8.1",
]

extras_require = {
    "boosted": [
        "xgboost",
        "lightgbm",
    ],
}
extras_require["complete"] = sorted({v for req in extras_require.values() for v in req})

# Only include pytest-runner in setup_requires if we're invoking tests
if {"pytest", "test", "ptr"}.intersection(sys.argv):
    setup_requires = ["pytest-runner"]
else:
    setup_requires = []

ADS_VERSION = "UNKNOWN"
with open(
    os.path.join(os.path.dirname(os.path.abspath(__file__)), "ads", "ads_version.json")
) as version_file:
    ADS_VERSION = json.load(version_file)["version"]

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="oracle_ads",
    version=ADS_VERSION,
    description="Oracle Accelerated Data Science SDK",
    author="Oracle Data Science",
    license="Universal Permissive License 1.0",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    url="https://docs.oracle.com/en-us/iaas/tools/ads-sdk/latest/index.html",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Universal Permissive License (UPL)",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    keywords="Oracle Cloud Infrastructure, OCI, Machine Learning, ML, Artificial Intelligence, AI, Data Science, Cloud, Oracle",
    include_package_data=True,
    install_requires=install_requires,
    python_requires=">=3.7",
    setup_requires=setup_requires,
    extras_require=extras_require,
    tests_require=[
        "pytest",
    ],
)



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2021 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

class Parser:
    @property
    def parts_of_speech(self):
        raise NotImplementedError()

    @property
    def nouns(self):
        raise NotImplementedError()

    @property
    def adjectives(self):
        raise NotImplementedError()

    @property
    def adverbs(self):
        raise NotImplementedError()

    @property
    def verbs(self):
        raise NotImplementedError()

    @property
    def words(self):
        raise NotImplementedError()

    @property
    def sentences(self):
        raise NotImplementedError()

    @property
    def histogram(self):
        raise NotImplementedError()

    @property
    def bigrams(self):
        raise NotImplementedError()

    @property
    def trigrams(self):
        raise NotImplementedError()



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2021 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

"""
The module that represents a Category feature type.

Classes:
    Category
        The Category feature type.
"""
import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns
from ads.feature_engineering.feature_type.base import FeatureType
from ads.feature_engineering.utils import (
    _count_unique_missing,
    _set_seaborn_theme,
    SchemeTeal,
)
from ads.feature_engineering import schema


class Category(FeatureType):
    """
    Type representing discrete unordered values.

    Attributes
    ----------
    description: str
        The feature type description.
    name: str
        The feature type name.
    warning: FeatureWarning
        Provides functionality to register warnings and invoke them.
    validator
        Provides functionality to register validators and invoke them.

    Methods
    --------
    feature_stat(x: pd.Series) -> pd.DataFrame
        Generates feature statistics.
    feature_plot(x: pd.Series) -> plt.Axes
        Shows the counts of observations in each categorical bin using bar chart.

    """

    description = "Type representing discrete unordered values."

    @staticmethod
    def feature_stat(x: pd.Series) -> pd.DataFrame:
        """Generates feature statistics.

        Feature statistics include (total)count, unique(count) and missing(count) if there are any.

        Parameters
        ----------
        x : :class:`pandas.Series`
            The feature being evaluated.

        Returns
        -------
        :class:`pandas.DataFrame`
            Summary statistics of the Series or Dataframe provided.

        Examples
        --------
        >>> cat = pd.Series(['S', 'C', 'S', 'S', 'S', 'Q', 'S', 'S', 'S', 'C', 'S', 'S', 'S',
                    'S', 'S', 'S', 'Q', 'S', 'S', '', np.NaN, None], name='сategory')
        >>> cat.ads.feature_type = ['сategory']
        >>> cat.ads.feature_stat()
            Metric  Value
        0	count	22
        1	unique	3
        2	missing	3
        """
        return _count_unique_missing(x)

    @staticmethod
    def feature_plot(x: pd.Series) -> plt.Axes:
        """
        Shows the counts of observations in each categorical bin using bar chart.

        Parameters
        ----------
        x : :class:`pandas.Series`
            The feature being evaluated.

        Returns
        -------
        matplotlib.axes._subplots.AxesSubplot
            Plot object for the series based on the Category feature type.

        Examples
        --------
        >>> cat = pd.Series(['S', 'C', 'S', 'S', 'S', 'Q', 'S', 'S', 'S', 'C', 'S', 'S', 'S',
                    'S', 'S', 'S', 'Q', 'S', 'S', '', np.NaN, None], name='сategory')
        >>> cat.ads.feature_type = ['сategory']
        >>> cat.ads.feature_plot()
        """
        col_name = x.name if x.name else "сategory"
        df = x.to_frame(name=col_name)
        df = df.dropna()
        if len(df.index):
            _set_seaborn_theme()
            ax = sns.countplot(y=col_name, data=df, color=SchemeTeal.AREA_DARK)
            ax.set(xlabel="Count")
            return ax

    @classmethod
    def feature_domain(cls, x: pd.Series) -> schema.Domain:
        """
        Generate the domain of the data of this feature type.

        Examples
        --------
        >>> cat = pd.Series(['S', 'C', 'S', 'S', 'S', 'Q', 'S', 'S', 'S', 'C', 'S', 'S', 'S',
                    'S', 'S', 'S', 'Q', 'S', 'S', '', np.NaN, None], name='category')
        >>> cat.ads.feature_type = ['category']
        >>> cat.ads.feature_domain()
        constraints:
        - constraint: in
            value:
            - S
            - C
            - Q
            - ''
            - .nan
            - null
        stats:
            count: 22
            missing: 3
            unique: 3
        values: Category

        Returns
        -------
        ads.feature_engineering.schema.Domain
            Domain based on the Category feature type.
        """

        return schema.Domain(
            cls.__name__,
            cls.feature_stat(x).to_dict()[x.name],
            [schema.Expression(f"$x in {x.dropna().unique().tolist()}")],
        )



"""Contains class wrapping OCI resource search service
"""
import oci
from ads.common.oci_mixin import OCIClientMixin


class SEARCH_TYPE(str):
    FREETEXT = "FreeText"
    STRUCTURED = "Structured"


class OCIResource(OCIClientMixin):
    """Contains helper methods for getting information from OCIResourceSearch service.

    Usage:
    Find the compartment ID of an OCI resource:
    >>> OCIResource.get_compartment_id("YOUR_OCID")
    Search for OCI resource matching free text (Similar to the search box in OCI console):
    >>> OCIResource.search("YOUR_FREE_TEXT")
    Search for OCI resource matching structured text:
    >>> OCIResource.search("STRUCTURED_TEXT", type="Structured")
    """

    @classmethod
    def init_client(cls, **kwargs) -> oci.resource_search.ResourceSearchClient:
        return cls._init_client(
            client=oci.resource_search.ResourceSearchClient, **kwargs
        )

    @classmethod
    def get_compartment_id(cls, ocid) -> str:
        """Gets the compartment OCID of an OCI resource, given the resource's OCID.

        Parameters
        ----------
        ocid : str
            OCID of a resource

        Returns
        -------
        str:
            Compartment OCID of the resource


        """
        results = set([r.compartment_id for r in cls.search(ocid)])
        if len(results) == 1:
            return list(results)[0]
        elif not results:
            raise Exception(f"Resource not found: OCID={ocid}")
        else:
            raise Exception(
                f"Unable determine compartment ID, multiple matches: {results}"
            )

    @classmethod
    def search(
        cls,
        query: str,
        type: str = "FreeText",
        config: dict = None,
        tenant_id: str = None,
        limit: int = 500,
        page: str = None,
        **kwargs,
    ) -> list:
        """Search OCI resource by free text.

        Parameters
        ----------
        query : str
            The content to search for.
        type :  str (optional)
            The type of SearchDetails, whether "FreeText" or "Structured".
            Defaults to "FreeText".
        config: dict (optional)
            Configuration keys and values as per SDK and Tool Configuration.
            The from_file() method can be used to load configuration from a file.
            Alternatively, a dict can be passed. You can validate_config the dict
            using validate_config(). Defaults to None.
        tenant_id: str (optional)
            The tenancy ID, which can be used to specify a different tenancy
            (for cross-tenancy authorization) when searching for resources in
            a different tenancy. Defaults to None.
        limit: int (optional)
            The maximum number of items to return. The value must be between
            1 and 1000. Defaults to 500.
        page: str (optional)
            The page at which to start retrieving results.

        Returns
        -------
        list
            A list of search results

        """
        if not config:
            client = cls.init_client()
        else:
            client = oci.resource_search.ResourceSearchClient(config=config)

        if type == "Structured":
            search_details = oci.resource_search.models.StructuredSearchDetails(
                type="Structured",
                query=query,
                matching_context_type=None,
            )
        elif type == "FreeText":
            search_details = oci.resource_search.models.FreeTextSearchDetails(
                text=query, type="FreeText"
            )
        else:
            raise ValueError(f"Invalid Type: {type}")

        return client.search_resources(
            search_details,
            limit=limit,
            page=page,
            tenant_id=tenant_id,
            **kwargs,
        ).data.items
